### A Pluto.jl notebook ###
# v0.19.11

using Markdown
using InteractiveUtils

# This Pluto notebook uses @bind for interactivity. When running this notebook outside of Pluto, the following 'mock version' of @bind gives bound variables a default value (instead of an error).
macro bind(def, element)
    quote
        local iv = try Base.loaded_modules[Base.PkgId(Base.UUID("6e696c72-6542-2067-7265-42206c756150"), "AbstractPlutoDingetjes")].Bonds.initial_value catch; b -> missing; end
        local el = $(esc(element))
        global $(esc(def)) = Core.applicable(Base.get, el) ? Base.get(el) : iv(el)
        el
    end
end

# ╔═╡ fec77c57-c21a-42c4-8e39-3c21c804504c
begin
	using StatsPlots, PlutoUI, Random, Statistics
	using Flux
	using MLDatasets
	import MLUtils: MLUtils, shuffleobs, eachobs, getobs
	using ProgressLogging
end

# ╔═╡ 5b40470e-6195-4307-8c78-993da5c79699
begin
	_check_complete(complete) = complete ? "✅" : "❌"
	
	md"""
	# Setup

	This section loads and installs all the packages. You should be setup already from assignment 1, but if not please read and follow the `instructions.md` for further details.
	"""
end

# ╔═╡ a983ed0e-fb18-4b21-985b-5214e09ee5f8
# Helper function
function eachbatch(data, batchsize) 
	eachobs(data, batchsize=batchsize)
end

# ╔═╡ 6d0284b7-1b4e-4241-a796-034c53b06fef
gr() # In this notebook we use the plotly backend for Plots.

# ╔═╡ 72fef447-cdf6-443a-ae6b-53c217602d98
md"""
### !!!IMPORTANT!!!

Insert your details below. You should see a green checkmark.
"""

# ╔═╡ 54ff2a52-13c5-4e55-ad57-7acf5474cae4
student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)

# ╔═╡ 5897039a-9571-4648-b1f6-a74efb6aef3b
let
	def_student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)
	if length(keys(def_student) ∩ keys(student)) != length(keys(def_student))
		md"You don't have all the right entries! Make sure you have `name`, `email`, `ccid`, `idnumber`. ❌"
	elseif any(getfield(def_student, k) == getfield(student, k) for k in keys(def_student))
		md"You haven't filled in all your details! ❌"
	elseif !all(typeof(getfield(def_student, k)) === typeof(getfield(student, k)) for k in keys(def_student))
		md"Your types seem to be off: `name::String`, `email::String`, `ccid::String`, `idnumber::Int`"
	else
		md"Welcome $(student.name)! ✅"
	end
end

# ╔═╡ c69ce5c8-cd42-4ae2-aaa9-74d0daa5875f
md"""
Important Note: You should only write code in the cells that have: """

# ╔═╡ dfb12ef6-2968-40bc-a7eb-5284dec0c359
#### BEGIN SOLUTION

#### END SOLUTION

# ╔═╡ 913c8492-d221-4cb1-aab9-cbb2dc2026f9
# This cell creates the table of contents.
TableOfContents(title="Assignment 4", depth=4)

# ╔═╡ 74bd4039-7a14-4c1e-95d3-e9f336293d09
md"""
# Some Assignment Tips
- The first question requires no implementation from you, but you can use the implementations to get an idea of how certain aspects of `Flux` works before moving on to Variational Autoencoders.
- There are checkboxes for the VAE section, to turn on `train VAE` or `train CVAE`. It is better to have these unchecked while making changes to the functions they rely on. The Pluto notebook forces every cell with a dependency on the changed cell to run, and so it will be needlessly running these when you are making changes. The unit tests are enough to see if you have things right, and you can run after you pass these.
- Note that, for this first part, the layer ordering is different from the one used in the notes (e.g. layers 1 means the input layer) 
"""

# ╔═╡ e1c3cf03-f235-4a91-aae0-7afce73272cb


# ╔═╡ 5b6b85d6-25f5-4ac9-9fe3-5e85c432f0f0
md"""
# Q3: How do ANNs represent and transform data?

Even though neural networks can be quite complex, we can still work to understand the models that they learn. In this section, we are going to do a simple two-dimensional example, based on [this nice blog post from Chris Olah](https://colah.github.io/posts/2014-03-NN-Manifolds-Topology/). 
"""

# ╔═╡ 6b7db8c9-6a3a-49ae-926a-35a8f8bfb8e8
md"""
### Generating the Data

First we need to generate data. We use two sinusoidal functions to generate data for two classes, visualized below. The inputs are two dimensional, $\mathbf{x} \in [-1,1]^2$, with all points on the blue curve corresponding to class 1 and all those on the red curve corresponding to class 2. These classes cannot be separated using a linear decision plane, but they can be perfectly separated with a nonlinear one. We will use this data to train a neural network using the cross-entropy. We expect that our neural network should be able to perfectly separate these classes. 

"""

# ╔═╡ 31543a74-a116-4861-b153-30e6bcaecce8
function class_1(x)
	1//2 * sin( (x-1//2) * π) + 2//5
end

# ╔═╡ b102d8ee-1191-4580-aa53-09740cf608a4
function class_2(x)
	1//2 * sin( (x-1//2) * π) - 2//5
end

# ╔═╡ f9596fd6-b1fa-43d4-b84c-9a1151db372d
function dataset_sin(n)
	X_c1 = reduce(hcat, 
		begin
			x=2*rand(Float32)-1
			y=class_1(x)
			[x, y]
		end for i in 1:n)
	
	
	X_c2 = reduce(hcat, 
		begin
			x=2*rand(Float32)-1
			y=class_2(x)
			[x, y]
		end for i in 1:n)
	
	Y = fill(false, 2*n)
	Y[1:n] .= true
	
	X = hcat(X_c1, X_c2)
	(X, reshape(Y, 1, :))
	
end

# ╔═╡ b37d3b82-fcc2-4ac9-94cc-86174481ba86
let
	plot(-1:0.01:1, class_1)
	plot!(-1:0.01:1, class_2, legend=nothing)
	X, Y = dataset_sin(30)
	scatter!(X[1, :], X[2, :], markercolor=reshape([y==1 ? :blue : :red  for y in Y'], :))
end

# ╔═╡ 0371d19e-9684-4b81-b1cb-91409fcab2d7
md"""
### Training a Simple Neural Network

We will use [Flux.jl](https://github.com/FluxML/Flux.jl), which is a neural network package written entirely in Julia. If you are interested in how to write more advanced julia packages, this is a great example package to learn and read. Flux is easily extensible.

In the following, we use [`Flux.train!`](https://fluxml.ai/Flux.jl/stable/training/training/) from Flux. We also use `eachbatch` and `shuffleobs` from `MLDataUtils` to create batches and appropriately shuffle the training data on each epoch. You will need these functions for the VAE section.
"""

# ╔═╡ fd07264f-a9c2-4dd7-83de-6fc6051986b4
function train!(loss, ps::Flux.Params, data, opt::Flux.Optimise.AbstractOptimiser)
	n = length(data)
	for (i, d) in enumerate(data)
		gs = gradient(ps) do
			loss(d...)
		end
		Flux.Optimise.update!(opt, ps, gs)
  	end
end

# ╔═╡ 226149d2-a4ce-4713-8ff4-420c82206b07
function train_classifier!(model; N_2=1024, batch_size=32, epochs=200, η=1e-2)
	
	D = dataset_sin(N_2)
	ps = Flux.params(model)
	opt = RMSProp(η)
	
	#= 
	Use Flux.Losses.binarycrossentropy to define the 
	loss function for the classifier.
	=#
	loss(x, y) = Flux.Losses.binarycrossentropy(model(x), y)

	#= 
	The training procedure using `train!`.
	=#
	for epoch in 1:epochs
		Ds = shuffleobs(D)
		train!(loss, ps, eachbatch(Ds, batch_size), opt)
	end

	model
end

# ╔═╡ 49288405-b370-4cfb-a0e3-01d3b9126470
md"""
#### Visualizing the Decision Surface

Spend some time with the following visualization. You can use the below interface to modify the activation functions, number of hidden layers, and number of hidden units in each layer. The input space of our problem is $\mathbf{x}\in[-1,1]^2$. We train a neural network using the binary cross entropy loss function to discriminate between class 1 and class 2. The plot below shows the output of the neural network, which is the estimate of $p(y = 1 | \mathbf{x})$. We explicitly outline the resulting decision plane, which is the line where $p(y = 1 | \mathbf{x}) = 0.5$.

You will need this visualization to answer the written part, and you will need to save plots as well. You can save the plot with the cell directly following the heatmap, or drag and drop the plot to your desktop (converting from `svg` to `pdf` after the fact). 

- Hidden Activation Function: $(@bind _hidden_act PlutoUI.Select(["sigmoid", "relu", "identity"]))
- Number of layers: $(@bind _layer_cnt_str PlutoUI.Select(string.(0:2), default="1"))
- Number of hidden units: $(@bind _hu_cnt_str PlutoUI.Select(string.([1, 2, 4, 8, 16]), default="2"))
"""

# ╔═╡ 8b85f477-674e-403e-a834-36eb53ab4c30
let

	_model = let
		Random.seed!(10)
		act = getfield(Flux, Symbol(_hidden_act))
		_layer_cnt = parse(Int, _layer_cnt_str)
		_hu_cnt = parse(Int, _hu_cnt_str)
		model = if _layer_cnt == 0
			Flux.Chain(Dense(2, 1, σ))
		else
			Flux.Chain([Dense(i == 1 ? 2 : _hu_cnt, _hu_cnt, act) for i in 1:_layer_cnt]..., 
					   Dense(_hu_cnt, 1, σ))
		end
		train_classifier!(model, epochs=200)
	
		model
	end;
	
	X_x = -1.0:0.01:1.0
	X_y = -1.0:0.01:1.0
	plt = heatmap(X_x, X_y, getindex.(_model.(collect.(Iterators.product(X_x, X_y))), 1)', title="Activation: $(_hidden_act), Layers: $(_layer_cnt_str), HUs: $(_hu_cnt_str)")
	plot!(plt, -1.0:0.01:1.0, class_1, legend=nothing)
	plot!(plt, -1.0:0.01:1.0, class_2, legend=nothing)
end

# ╔═╡ df07dbe6-3983-48a8-b92a-40ccd4f08a87
# Manually uncomment and run this cell, you can also drag and drop:
# savefig("location.pdf")

# ╔═╡ 54f27f56-945a-48ae-96cf-51e15ca27cbe
md"""
### Visualizing Hidden Unit Activations

In this section we visualize the internal representation of some of the networks in the above example. 

"""

# ╔═╡ f8f90dbf-3b35-4ec3-8b65-1652aab3f954
md"""
First, we need to build up the machinery to capture the representation of the internal hidden units of the network. Flux comes to the rescue with the function `Flux.activations`. This returns all the activations of the network given an input. 
"""

# ╔═╡ c4df8697-e133-4e92-b271-4ba5371566fa
#=
This cell creates a new model `test_model` which we will use for some visualizations and testing below.
=#
test_model = let
	Random.seed!(12)
	act = relu
	_layer_cnt = 3
	_hu_cnt = 4
	model = if _layer_cnt == 0
		Flux.Chain(Dense(2, 1, σ))
	else
		Flux.Chain([Dense(i == 1 ? 2 : _hu_cnt, _hu_cnt, act) for i in 1:_layer_cnt]..., 
				   Dense(_hu_cnt, 1, σ))
	end
	train_classifier!(model, epochs=200)

	model
end;

# ╔═╡ 697d1728-1ec3-4a76-9b9d-de2831e39214
Flux.activations(test_model, [-1.0, 0.44])

# ╔═╡ bd1e57af-6373-4815-a305-86e03f57c64e
function get_activation_index(acts, layer, hidden_unit)
	# This should be a single line
	acts[layer][hidden_unit]
end

# ╔═╡ 3d49ff71-63cf-495e-b87c-e0f3ba7b925d
begin
	__get_act_index_check = let
		acts = [[1, 2, 3], [4, 5, 6], [7, 8]]
		c1 = get_activation_index(acts, 1, 2) == 2
		c2 = get_activation_index(acts, 2, 3) == 6
		c3 = get_activation_index(acts, 3, 1) == 7
		c1 && c2 && c3
	end
	
md"""
The `get_activation_index` function takes the return of `Flux.activations`, a layer index, and a hidden_unit index. This should return the activation for that layer and hidden unit.
"""
end

# ╔═╡ 7b6fc4a5-a630-42e7-9924-c2862bebfdd1
md"""
The `plot_activations` function creates a heatmap showing the activation of one hidden unit across the 2d input space. We query the network for an input $\mathbf{x}\in[-1, 1]^2$ and record the activation for the hidden unit. We repeat this many times over a fine-grained grid of $\mathbf{x}\in[-1, 1]^2$), to get a heatmap over the 2d space.
"""

# ╔═╡ 356d6228-0b23-441f-826a-ac8b1cf495e8
function plot_activations(model, layer, hidden_unit)
	X_x = -1.0:0.01:1.0
	X_y = -1.0:0.01:1.0
	
	acts = Flux.activations.((model,), collect.(Iterators.product(X_x, X_y)))
	heatmap(X_x, X_y, get_activation_index.(acts, layer, hidden_unit)')
	plot!(-1.0:0.01:1.0, class_1, legend=nothing)
	plot!(-1.0:0.01:1.0, class_2, legend=nothing)
end

# ╔═╡ f64a54e5-930d-46fd-9741-6cd87696b2e7
md"""
Here is an example visualizing a hidden unit's activation across inputs $\mathbf{x}\in[-1, 1]^2$. By default, this shows the first hidden unit in layer 1, for a network with three hidden layers with four nodes per hidden layer. You can change which layer and hidden unit you look at. The lines overlayed on the heatmap correspond to the sinusoidal functions used to generate the training data. Note that we can query the network for any $\mathbf{x}\in[-1, 1]^2$, even though the network was only trained on the subset of $\mathbf{x}$ that fall on this lines.

- Layer: $(@bind _lay_str PlutoUI.Select(string.([1, 2, 3]), default="1"))
- Hidden Unit: $(@bind _hu_str PlutoUI.Select(string.(1:4), default="1"))
"""

# ╔═╡ 36fd8579-69bb-4e6d-98a3-5e80e2a2f3b9
plot_activations(test_model, parse(Int, _lay_str), parse(Int, _hu_str))

# ╔═╡ 5d5019a1-42fa-4076-9de1-71c575133f50
md"""
### Data Transformations

We can also look at how different parts of the network transform the input data and how that relates to the final decision plane. The function `plot_transform` let's us see how the hidden nodes relate to the output. We can think of this as first mapping $\mathbf{x}$ using the first layers of the neural network, to get a new representation $\boldsymbol{\phi}(\mathbf{x})$ and then visualizing the decision surface in this new space. That is, visualizing the decision surface in terms of inputs $\boldsymbol{\phi}(\mathbf{x})$. 

The function `plot_transform` takes a model and a layer `l::Int`. This function assumes each layer has no more than 2 hidden units and that the activation is a sigmoid. We use 2 hidden units to make it easy to visualize. The neural network can be of any depth. The generated heatmap has lines for the data generating functions, like in the plots above, though now the lines have a slightly different interpretation.

- **Heatmap**: The heatmap is the output of the neural network, now with respect to $\phi_1$ and $\phi_2$ on the two axes. It is generated by varying $\phi_1 \in [0,1]$ and $\phi_2 \in [0,1]$, where the range is $[0,1]$ because we assumed a sigmoid activation. 
- **Lines**: The lines are generated by taking $\mathbf{x}$ from the sinusoidal data generating function and plotting $\boldsymbol{\phi}(\mathbf{x})$. It let's us visualize the data in the new space. If $\boldsymbol{\phi}(\mathbf{x})$ is the last hidden layer, then it shows us the space in which we learn a linear classifier and lets us see if the data has been transformed such that it is now linearly separable.

"""

# ╔═╡ b5cbed07-f2f0-4360-9714-5a7218831e55
function plot_transform(model, layer; kwargs...)
	
	i = layer
	front_model = model[1:i]
	rear_model = model[(i+1):end]
	
	X_x = 0.0:0.01:1.0
	X_y = 0.0:0.01:1.0
	
	heatmap(X_x, X_y, getindex.(rear_model.(collect.(Iterators.product(X_x, X_y))), 1)'; kwargs...)
	
	X_x = -1.0:0.01:1.0
	X_y = class_1.(X_x)
	
	plot!(getindex.(front_model.(collect.(zip(X_x, X_y))), 1), getindex.(front_model.(collect.(zip(X_x, X_y))), 2), legend=nothing, lw=3)
	
	X_x = -1.0:0.01:1.0
	X_y = class_2.(X_x)
	plot!(getindex.(front_model.(collect.(zip(X_x, X_y))), 1), getindex.(front_model.(collect.(zip(X_x, X_y))), 2), legend=nothing, lw=3)

end

# ╔═╡ a534dcbd-33f4-4920-950b-ae3f34142bd8
md"""
#### What happens when a model fails?

Let's plot the decision surface for a run where the learned neural network model was successful and a run where it failed and the model was poor. The plots below make it clear why the failed model classifies so poorly; you will explain this in the written part of your assignment.
"""

# ╔═╡ c7f2fc33-4c0d-402a-a8bf-d686458d1b67
md"""
Below is a succesful run, producing a good model.
"""

# ╔═╡ 0390b28b-0b50-424c-8f56-4c2dd01afdc5
successful_model = let
	Random.seed!(10)
	act = sigmoid
	model = Flux.Chain(
		Dense(2, 2, act),
		Dense(2, 2, act),
		Dense(2, 1, σ))
	train_classifier!(model)
	
	model
end

# ╔═╡ 46d0b27d-4cb2-4916-a226-9ab34e561cd9
plot_transform(successful_model, 2)

# ╔═╡ eb82338f-bfc3-4ab8-a30a-abf2ec333593
md"""

Below is a failed run, producing a poor model with low accuracy.
"""

# ╔═╡ 299c202e-b1a9-4e90-8efc-b15fb44c8e5e
failed_2layer_model = let
	Random.seed!(11)
	act = sigmoid
	model = Flux.Chain(
		Dense(2, 2, act),
		Dense(2, 2, act),
		Dense(2, 1, σ))
	train_classifier!(model)
	
	model
end

# ╔═╡ 6b366eac-6b94-44a8-969a-1148577b251d
plot_transform(failed_2layer_model, 2)

# ╔═╡ 49f77c7b-6e6d-4cf9-a0ea-b90c99891570
md"""
#### Epoch by Epoch

Below we look at the data transformations after each epoch of training for a successful run. Again, you will use these plots to answer questions in the written part of the assignment.

**Note**: On first load, your browser may not show the entire heatmap. You can sweep over all 30 images and have them all load, and they should fully load in a few seconds. You can also try resizing your window to force the browser to redraw the screen.


Epoch: $(@bind plt_idx PlutoUI.NumberField(0:30))
"""

# ╔═╡ 390fc91e-0fb9-4f32-9a3a-54c4df1954bf
rep_plots = let
	Random.seed!(10)
	model = Flux.Chain(Dense(2, 2, sigmoid), Dense(2, 2, sigmoid), Dense(2, 1, σ))
	D = dataset_sin(1024)
	ps = Flux.params(model)
	opt = RMSProp(1e-2)
	title_plt = plot(title = "Epoch-0", grid = false, showaxis = false, bottom_margin = -20Plots.px, ticks=false)
	plts = [plot!(title_plt, plot_transform(model, 1, title="Layer-1"), plot_transform(model, 2, title="Layer-2",), layout=@layout([A{0.01h}; [B C]]))]
	for e in 1:30
		Ds = shuffleobs(D)
		train!((x, y) -> Flux.Losses.binarycrossentropy(model(x), y), ps, eachbatch(Ds, 16), opt)
		title_plt = plot(title = "Epoch-$(e)", grid = false, showaxis = false, bottom_margin = -20Plots.px, ticks=false)
		push!(plts, plot(title_plt, plot_transform(model, 1, title="Layer-1"), plot_transform(model, 2, title="Layer-2"), layout=@layout([A{0.01h}; [B C]])))
	end
	plts
end;

# ╔═╡ 09d8ef7a-b96c-4f50-a470-acbbc2d5b895
rep_plots[plt_idx+1]

# ╔═╡ 5b6519a7-00dd-4e18-a8e9-a9eaed60d60e
md"""
# Variational Autoencoders

In the following section, we will be implementing the details of a variational autoencoder. Below you will find two structs. One for the `Encoder` and another for the full `VAE`. We will be using Flux's built in utilites for the `Decoder` network.

**IMPORTANT:** Unlike in previous assignments, we are breaking with some notational assumptions we made in the notes. The matrix of instances `X` will be constructed where each column is an instance instead of each row. This means $X \in \mathbb{R}^{\text{\# Features} \times \text{\# Samples}}$. We are breaking this assumption to match with Flux’s and Julia’s ecosystem.
"""

# ╔═╡ 38f9f22f-fabf-4278-9a5a-2f77eac7c2db
function encode_input(encoder, x)
	f = encoder.encode_feats(x)
	encoder.μ(f), encoder.logvar(f)
end

# ╔═╡ 8c41e1cf-8139-4e61-b92b-b7d201e91d2d
begin
	struct Encoder{F, M}
		encode_feats::F
		μ::M
		logvar::M
		function Encoder(ef_model::F, ef_out, latent) where F
			μ, logvar = Dense(ef_out, latent), Dense(ef_out, latent)
			new{F, typeof(μ)}(ef_model, μ, logvar)
		end
	end
	Flux.@functor Encoder
	function (encoder::Encoder)(x)
		encode_input(encoder, x)
	end
end

# ╔═╡ d5fa8fc7-ddd0-4739-8ed2-d3c45995e179
begin
	struct VAE
		encoder::Encoder
		decoder::Flux.Chain
	end
	Flux.@functor VAE # to make the struct work w/ Flux's autodiff
end

# ╔═╡ 32bfe0e2-21f0-4d5e-9c60-2a2ed5ca31c3
function reconstruct(vae::VAE, x)
	μ, logvar = vae.encoder(x)
	## TODO: Get the correct h for the decoder
	#### BEGIN SOLUTION
	zeros(Float32, size(vae.encoder.μ.weight, 1), size(x, 2))

	#### END SOLUTION
	μ, logvar, vae.decoder(h)
end

# ╔═╡ 4734248c-4542-4531-9b43-ee14aa10f306
function kl_divergence(μ, logvar)
	#### BEGIN SOLUTION
	0.0f0

	#### END SOLUTION
end

# ╔═╡ 17db8cf5-9c65-4445-b629-4b96dbd3a319
md"Below is the check for the function `kl_divergence`"

# ╔═╡ 0df361d4-5a02-4445-b792-776080ec0ae6
__check_kl_div = let 
	rng = Random.Xoshiro(1)
	v = kl_divergence(randn(rng, Float32, 2, 10), randn(rng, Float32, 2, 10))
	v isa Float32 && v ≈ 1.666146f0
end

# ╔═╡ e6283131-2e8b-453f-867c-56cd91a48527
let
	
	__int_kl_div = if @isdefined __check_kl_div
		__check_kl_div
	else
		false
	end

md"""
#### KL Divergence $(_check_complete(__int_kl_div))

We need to estimate the KL divergence between our variational distribution $q(\mathbf{h} | \mathbf{x})$ and the prior over $h$, $p(\mathbf{h})$. For the VAE, $p(\mathbf{h})$ is a zero-mean unit-variance Gaussian distribution, so this KL divergence simplifies to 

```math
D_{KL}(q(\cdot | \mathbf{x}_i) || p(\cdot || \mathbf{x}_i, \mathbf{W})) = \frac{1}{2} \sum_{j} \mu_j^2 + \sigma^2_j - \ln(\sigma^2_j) - 1
```

The above equation is for a single instance $\mathbf{x}_i$. In the below function, you should implement this equation for a batch of instances, where you take the average KL divergence over the batch.

"""
end

# ╔═╡ e417b47c-9f48-4d3c-82be-0a890c306266
function binary_cross_entropy(x̂, x)
	#### BEGIN SOLUTION
	0.0f0

	#### END SOLUTION
end

# ╔═╡ b76e5669-7ed6-4006-9007-934136bd9350
__check_bce = let
	rng = Random.MersenneTwister(1)
	x, x̂ = Float32.(rand(rng, Bool, 28, 32)), Float32.(rand(rng, Float32, 28, 32))
	binary_cross_entropy(x̂, x) ≈ 28.755898f0
end

# ╔═╡ fc14bedb-e3df-4c2d-9902-1b9621e372be
let
	
	__int_bce = if @isdefined __check_bce
		__check_bce
	else
		false
	end

	md"""
	#### Binary Cross Entropy $(_check_complete(__int_bce))
	
	We want to minimize the negative log-likelihood of our reconstruction, $-\mathbb{E}[\ln p(\mathbf{x} | \mathbf{h}, \mathbf{W})] $, approximated with a mini-batch of size $b$
	
	```math
	\frac{1}{b}\sum^b_{i=1} -\ln p(\mathbf{x}_i | \mathbf{h}_i, \mathbf{W})
	```
	
	where $b$ is the batch size. We will be using the MNIST data set. We assume that each pixel in the MNIST data set is a 0 or 1, and we model the distribution over each pixel as a Bernoulli distribution. Thus each term in this sum is 
	
	```math
	-\ln p(\mathbf{x}_i | \mathbf{h}_i, \mathbf{W}) = \sum_{j} -x_{ij} \ln(\hat{x}_{ij}) - (1-x_{ij}) \ln(1 - \hat{x}_{ij})
	```
	
	where $\hat{\mathbf{x}}_i = \text{Decoder}(\mathbf{h}_i)$ and $x_{ij}$ is the index $j$ into the vector $\mathbf{x}_i$ which corresponds to instance $i$.
	
	"""
end

# ╔═╡ a1e5e164-68f7-4838-8a06-d268c263fdc7
md"Below is the test for the `elbo` function."

# ╔═╡ dab0476f-8da9-4d6e-a418-9c4e0eda9e4a
md"""
### VAE on MNIST 

Below we build up the necessary utilities to train our VAE on the MNIST dataset.

The following two functions get the data from MNIST (saving temporary files in a folder called "/tmp"). We then bin the data set so that any pixel $>0.5$ is white and everything else is black. This makes sure the dataset actually follows our assumption that elements in $\mathbf{x}$ are binary.
"""

# ╔═╡ b1a674fa-8893-4b6a-b38e-a71474165786
function get_mnist_data()
	X,Y = (MLDatasets.MNIST(split=:train, dir="./tmp").features, 	
			MLDatasets.MNIST(split=:train, dir="./tmp").targets)
    X = Float32.(reshape(X, 28^2, :) .> 0.5)
	X, Y
end

# ╔═╡ 0616c043-10e3-4d1e-8328-7a694a931dce
function get_mnist_test_data()
    X, Y = (MLDatasets.MNIST(split=:test, dir="./tmp").features, 	
			MLDatasets.MNIST(split=:test, dir="./tmp").targets)
    X = Float32.(reshape(X, 28^2, :) .> 0.5)
	X, Y
end

# ╔═╡ b3d2988a-d99b-4b16-bace-31f0134ed36b
md"""
Below are the arguments we use for running the VAE experiments. These are shared for both the VAE and CVAE experiments.
"""

# ╔═╡ e32e185d-bfc9-411a-ade0-c6cb843ef0d6
Base.@kwdef struct MNIST_Args
	batch_size = 128        # batch size
	epochs = 20             # number of epochs

	# optimizer
	η = 1e-3                # learning rate

	# network
	input_dim = 28^2        # image size
	latent_dim = 16         # latent dimension
	hidden_dim = 64         # hidden dimension
end

# ╔═╡ d1fd5e60-bb94-447e-9b35-39ef07d328b9
md"""
Below is the function we will use to create and train a VAE on the MNIST dataset. You need to complete two tasks for this function to work.

**Create the VAE model:**

Using the VAE struct and Encoder struct you need to construct the VAE we are going to train. The shared portion of the Encoder will be a single layer with `tanh` activations and an output of `args.hidden_dim` (as well as passing in the other required arguments). The decoder will be a `Flux.Chain` with two layers and activations `tanh` and `sigmoid` respectively. The first layer will have an input the size of our latent state with `args.hidden_dim` as the output. The last layer will have an input of size `args.hidden_dim` and an output the size of our input.

**Note:** `tanh` and `sigmoid` are provided and don't need to be implemented. The functions just need to be passed as activations to the appropriate layers. 

**Implement part of the training loop:**

Because our `elbo` loss does not take a label, using the training loop defined by flux is not exactly what we need. Instead, we can write our own training loop! We can two functions provided by `Flux` to accomplish this:

- `gradient(loss, ps::Flux.Params)`: takes in a set of parameters to train and returns an `IdDict` contiaining the gradients of these parameters wrt the loss function. The loss function is a function which takes no arguments, you should create a lambda _or_ use a do block to pass this in.

- `Flux.Optimise.update!(opt, ps, grads)` this actually updates our model parameters in `ps=Flux.params(model)` with our optimizer `opt` according to the gradients `grads`.

The inner portion of your training loop should look like:
```julia
train_data, opt, model = #Defined somewhere
ps = Flux.params(model)

### Loops
grads = gradient(ps) do
	# batch[1] are the features, batch[2] are the labels.
	loss(model, batch[1], batch[2]) 
end
Flux.Optimise.update!(opt, ps, grads)
### End Loops
```

We implement part of this for you below, where we loop over `args.epochs` and batches of size `args.batch_size`. Your job is to use the correct loss and model. 

Note that if you go look at the code for [`Flux.train!`](https://github.com/FluxML/Flux.jl/blob/ea26f45a1f4e93d91b1e8942c807f8bf229d5775/src/optimise/train.jl#L99) you will find that it is remarkably simple. It can help you understand what you need to accomplish. 
"""

# ╔═╡ 0ac7b433-e358-4bb8-99de-bf43d6a5cf23
md"""
**Tip:** You can use the macro `@progress` to create a progress bar for loops. 
"""

# ╔═╡ de0817a7-0106-416e-a8de-dddcc99c3a5e
@progress for i in 1:10
	sleep(0.1)
end 

# ╔═╡ 529abd48-aa20-445d-ba5d-44b31749de70
begin

md"""
## Conditional VAE

Now we extend the above to allow us to condition on other variables, to get a conditional VAE. For MNIST, for example, instead of generating any image $p(\mathbf{x})$ we might want to generate only ones by sampling from $p(\mathbf{x} | y = 1)$ or only fives by sampling from $p(\mathbf{x} | y = 5)$. 

It is actually surprisingly straightforward to do this extension! Now instead of $q(\mathbf{h} | \mathbf{x})$ we learn $q(\mathbf{h} | \mathbf{x}, y)$. This involves simply including $y$ in the inputs to the encoder, since recall that the encoder is exactly learning this sampling distribution over $\mathbf{h}$ conditioned on inputs. Now those inputs are simply both $\mathbf{x}$ and $y$. 

The other change is that the decoder needs to condition on $y$. In other words, we need to learn $p(\mathbf{x} | \mathbf{h}, y)$ instead of $p(\mathbf{x} | \mathbf{h})$. Again, this just entails passing both $y$ and $\mathbf{h}$ into the decoder. During training, the $y$ given to the decoder should match the label for our image. When sampling from the decoder after training, we get to pick $y$ based on which images we want to see.  

In summary, the main difference between a conditional VAE (CVAE) and a vanilla VAE is in the input (i.e. what our function is conditioned on) for both the encoder and decoder. The CVAE will take an extra set of inputs (concatenated to the image or latent state), which is a one-hot encoding of the label.
"""
end

# ╔═╡ 25215b53-9a20-4cbc-bcee-17ed057950aa
begin
	struct CVAE
		encoder::Encoder
		decoder::Flux.Chain
	end
	Flux.@functor CVAE # to make the struct work w/ Flux's autodiff
end

# ╔═╡ dec6630f-504b-44dd-9e0a-0f9e796691ff
function append_class_to_data(X, Y)
	bit_Y = zeros(Float32, 10, length(Y))
	Flux.Zygote.ignore() do  # Ensures autograd won't backprop through this operation
		for i in 1:length(Y)
			bit_Y[Y[i] + 1, i] = 1
		end
	end
	vcat(X, bit_Y)
end

# ╔═╡ de0ba0ba-c061-4cae-8198-79524d892487
function reconstruct(cvae::CVAE, X, Y)
	X_n = append_class_to_data(X, Y)
	#### BEGIN SOLUTION
	zeros(Float32, size(cvae.encoder.μ.weight, 1), size(X, 2)), zeros(Float32, size(cvae.encoder.logvar.weight, 1), size(X, 2)), zero(X)

	#### END SOLUTION
end

# ╔═╡ f43279f9-6d3c-4d16-bb31-baddb402b54c
__check_recon_vae = let
	Random.seed!(1)
	encoder = Encoder(
		Dense(1, 2, tanh), 
		2, 
		2)
	decoder = Flux.Chain(
		Dense(2, 1, tanh), 
		Dense(1, 1, σ))
	vae = VAE(encoder, decoder)
	μ, logvar, x̂ = reconstruct(vae, rand(Float32, 1, 6))

	# check types of returned variables
	c_type = eltype(μ) <: Float32 && eltype(logvar) <: Float32 && eltype(x̂) <: Float32
	# check values of returned variables.
	c_val = μ[2, 4] ≈ -0.07837159f0 && logvar[1, 6] ≈ 0.052059837f0 && x̂[1, 5] ≈ 0.4106495f0
	c_type && c_val
end

# ╔═╡ f11e0cd6-ccd9-415e-9a20-a336227945f1
let

	__int_recon_vae = if @isdefined __check_recon_vae
		__check_recon_vae
	else
		false
	end
	
	md"""
	### Reconstructing the Input $(_check_complete(__int_recon_vae))
	
	We first setup the code to input $\mathbf{x}$ into the encoder, and sample $\mathbf{h}$, to pass to the decoder. This procedure can be seen like reconstructing $\mathbf{x}$ (and so is the terminology we use). It actually consists of two key parts. The encoder helps us learn $q(\mathbf{h} | \mathbf{x})$ so that we can get a better sampler for $\mathbf{h}$ in our optimization. This part outputs the means and variances for $q(\mathbf{h} | \mathbf{x})$. Learning this encoder is key for our optimization but is not strictly needed when we use the model. 
	
	The decoder is actually our final model that we use for sampling. When using it, we will query it with inputs $\mathbf{h} \sim \mathcal{N}(\mathbf{0}, \mathbf{I})$ to get samples of $\mathbf{x}$. When optimizing the decoder, however, we use samples from $q(\mathbf{h} | \mathbf{x})$. The decoder outputs $\hat{\mathbf{x}}$, the reconstruction that we want to make similar to $\mathbf{x}$. In our implementation, we optimize the encoder and decoder jointly, with one loss. `reconstruct`, therefore, outputs all the terms we need for this one loss.
	
	The encoder outputs means and log-variances, one for each dimension of $\mathbf{h}$. As in the notes, we use a log-variance, called logvar, to parameterize the Gaussian $\mathcal{N}\left(\mu, e^{logvar}\right)$. This helps with numerical stability, and means we can use a linear layer for estimating the variance. If we did not learn the $logvar = \log(\sigma^2)$ and use $e^{logvar}$, then we would have to ensure the variance output from the network is positive in some other way.
	
	As a useful note for this implementation, recall that if you have a standard normal random variable $Z \sim \mathcal{N}(0, 1)$ you can change its mean and variance to $\mu, \sigma^2$ respectively by multiplying by $\sigma$ and adding $\mu$ like this: $\mu + \sigma Z$. Hence, if $Y = \mu + \sigma Z$, then $Y \sim \mathcal{N}(\mu, \sigma)$.
		
	Below, you should implement a function which takes in some imput $\mathbf x$, predicts $\mathbf h$, and then reconstructs $\hat{\mathbf{x}} \approx \mathbf x$. We highly recommend that you (a) use `1//2` or `0.5f0` for the exponent, and (b) use `randn(Float32, size...)` for sampling random normal numbers. This way we can ensure the types are stable throughout the reconstruction; this matters a lot for performance in Flux which uses `Float32`s by default. If you used a different kind of number, Flux will have to promote it, which can be expensive.
	"""
end

# ╔═╡ 91aad47c-b715-42b2-8f79-074b016356c3
function elbo(vae, x)
	μ, logvar, x̂ = reconstruct(vae, x)

	# Ensure each output is a Float32
	@assert eltype(μ) <: Float32
	@assert eltype(logvar) <: Float32
	@assert eltype(x̂) <: Float32
	
    len = size(x)[end]
	
    # KL-divergence
    kl_q_p = kl_divergence(μ, logvar)::Float32
	
	# cross entropy
    logp_x_z = binary_cross_entropy(x̂, x)::Float32
	
	# return the final loss of elbo.
	#### BEGIN SOLUTION
	0.0f0

	#### END SOLUTION
end

# ╔═╡ 557c4bc1-74e7-409a-b620-2bb6592c5ee9
__check_elbo = let
	Random.seed!(1)
	encoder = Encoder(
		Dense(1, 2, tanh), 
		2, 
		2)
	decoder = Flux.Chain(
		Dense(2, 1, tanh), 
		Dense(1, 1, σ))
	vae = VAE(encoder, decoder)

	elbo(vae, Float32.(rand(Bool, 1, 32))) ≈ 1.8917041f0
end

# ╔═╡ 131a7831-250a-47e0-90f7-9cb14db0a53b
let
	__int_elbo = if @isdefined __check_elbo
		__check_elbo
	else
		false
	end

	md"""
	### Elbo Loss $(_check_complete(__int_elbo))
	
	As you've read in your notes the elbo loss has two parts. You will need to implement both the `kl_divergence` and the `binary_cross_entropy`. The `binary_cross_entropy` consists of the negative log-likelihood terms $-\ln p(\mathbf{x}_i | \mathbf{h}_i, \mathbf{W})$. You will use these functions to define the `elbo` loss used to train our VAE.
	
	**Warning:** Like above, you should be careful about how you implement these losses. You want to ensure you are not doing any implicit promotions to double precision floating point numbers (i.e. `Float64`). This is important for performance when training your network. A promotion will happen whenever two numbers of different types interact, leading to the number with higher precision to be chosen. For example if you type in `1 + 1.0f0` you will get `2.0f0` which is a `Float32`. You should also use vectorized implementations instead of raw loops (see [this tutorial for tips](https://mkschleg.github.io/tutorials/LAtoLoops.html)).
	"""
end

# ╔═╡ 7402bdc2-eb6f-4837-96f9-2d113d826b55
begin
	__int_elbo = if @isdefined __check_elbo
		__check_elbo
	else
		false
	end

md"""
#### elbo $(_check_complete(__int_elbo))

Write the output of the elbo loss function below.
"""
end

# ╔═╡ 60abff0f-1a7a-47a4-94b3-880a5894d139
function train_mnist_vae(; kwargs...)

	args = MNIST_Args(;kwargs...)
	
	Random.seed!(1)

	train_data = get_mnist_data()

	# Construct the VAE model
	#### BEGIN SOLUTION
	vae = nothing

	#### END SOLUTION
	
	ps = Flux.params(vae)
	
	opt = ADAM(args.η)
	
	# The training loops over epochs and batches.
	@progress for epoch = 1:args.epochs
		train_data_s = shuffleobs(train_data)
		for batch in eachbatch(train_data_s, args.batch_size)
			# Compute the gradient 
			#### BEGIN SOLUTION			

			#### END SOLUTION
			end
			Flux.Optimise.update!(opt, ps, grads)
		end
	end
	
	vae
end

# ╔═╡ ef533aa8-eccb-46a4-8ff9-ca49a233a46e
__check_cvae_recon = let
	Random.seed!(1)
	encoder = Encoder(
		Dense(1 + 10, 2, tanh), 
		2,
		2)
	decoder = Flux.Chain(
		Dense(2 + 10, 2, tanh), 
		Dense(2, 1, σ))
	cvae = CVAE(encoder, decoder)
	μ, logvar, x̂ = reconstruct(cvae, rand(Float32, 1, 16), rand(0:9, 1, 16))
	μ[1, 2] ≈ -0.33862936f0 && 
	logvar[2, 10] ≈ -0.35973492f0 && 
	x̂[1, 16] ≈ 0.3853241f0
end

# ╔═╡ a70afce3-8dab-4b1b-a09b-c81cc45cd09d
let
	__int_cvae_recon = if @isdefined __check_cvae_recon
		__check_cvae_recon
	else
		false
	end
	md"""
	### Reconstruct $(_check_complete(__check_cvae_recon))
	
	You will need to write the `reconstruct` $(_check_complete(__check_cvae_recon)) function for the CVAE. This should be very similar to the `VAE` version but using the `append_class_to_data` function to append the appropriate label to both the image data `X` and to the latent state `h`. We have done this for you for `X` so you can see how to use this function.
	"""
end

# ╔═╡ 86b19883-6574-4f10-a822-4186c4c0235c
md"""
Below you will implement `celbo`  which is the conditional elbo loss you derived in the written portion. This should look very similar to the above `elbo` loss and doesn't require any new helper functions (i.e. you should re-use `kl_divergence` and `binary_cross_entropy`).
"""

# ╔═╡ 33eacfac-60bd-4b89-afd3-1e92b61cf55d
function celbo(cvae::CVAE, X, Y)
	μ, logvar, X̂ = reconstruct(cvae, X, Y)
	@assert eltype(μ) <: Float32
	@assert eltype(logvar) <: Float32
	@assert eltype(X̂) <: Float32
	len = size(X)[end]
	
	#### BEGIN SOLUTION
	0.0f0

	#### END SOLUTION
end

# ╔═╡ 4c5d66c1-aa1e-49e9-b004-ac66fbc783c4
__check_celbo = let
	Random.seed!(1)
	encoder = Encoder(
		Dense(1 + 10, 2, tanh), 
		2,
		2)
	decoder = Flux.Chain(
		Dense(2 + 10, 2, tanh), 
		Dense(2, 1, σ))
	cvae = CVAE(encoder, decoder)
	celbo(cvae, Float32.(rand(Bool, 1, 16)), rand(0:9, 1, 16)) ≈ 0.9674432f0
end

# ╔═╡ 213781b6-0ca3-4439-908d-b1c1895be24e
let
	__int_celbo = if @isdefined __check_celbo
		__check_celbo
	else
		false
	end
	md"""
	### Conditional Elbo $(_check_complete(__check_celbo))
	"""
end

# ╔═╡ 64fa9dbe-f9f4-407d-8e19-02831a097eb6
md"""
Below you will implement the training function. This should look very similar to `train_mnist_vae` with some minor differences. Specifically,
- the loss function will be different
- the CVAE's network will be almost exactly the same as the VAE above except the input to the encoder and decoder will be slightly different. (hint: think about the `reconstruct` for CVAE you just implemented and remember mnist has 10 classes).
"""

# ╔═╡ e31da691-aad0-496c-addc-6730763ee7bb
function train_mnist_cvae(; kwargs...)

	args = MNIST_Args(;kwargs...)
	
	Random.seed!(1)

	train_data = get_mnist_data()
	
	@assert train_data[1] isa Matrix{Float32}

	# Construct the VAE
	#### BEGIN SOLUTION
	cvae = nothing

	#### END SOLUTION
	
	ps = Flux.params(cvae)
	
	opt = ADAM(args.η)
	
	# Write the training loops over epochs and batches.
	#### BEGIN SOLUTION

	#### END SOLUTION
	cvae
end

# ╔═╡ 5497634b-361e-4ad1-b800-215b91426af2
md"""
### Results

Below, we compare the output of a VAE with that of a Conditional VAE. Training each VAE takes approximately 2.5 minutes using a single core.

Train VAE: $(@bind _check_train_vae PlutoUI.CheckBox()) \
Train CVAE: $(@bind _check_train_cvae PlutoUI.CheckBox())
"""

# ╔═╡ acea195c-0075-4217-8235-6e444d2024a8
trained_vae = _check_train_vae ? train_mnist_vae() : nothing

# ╔═╡ c0eff92e-b378-4da1-b44a-02fce5095612
trained_cvae = _check_train_cvae ? train_mnist_cvae() : nothing

# ╔═╡ 08cc03f4-9338-422f-9884-aa2745d67812
md"""
Below we visualize the reconstruction of the two VAEs. If the image is black, then this means you have not trained the VAE for that label. The left most column is the ground truth from the dataset. The middle column is the vanilla VAE, and the right most column is the Conditional VAE. Both models are trained using 20 epochs, and the default parameters defined in `MNIST_Args`.
"""

# ╔═╡ e6cdb6d4-6c4a-473b-85cc-b45f7de2409c
(mnist_data_X, mnist_data_Y) = get_mnist_test_data();

# ╔═╡ 3b720427-6fb8-46e9-8cbe-93d69c0b268d
let
	indicies = [10, 9392, 232, 1, 132]
	plts = []
	Random.seed!(1)
	for i ∈ indicies
		
		p̂_vae = if trained_vae isa Nothing
			zero(getobs(mnist_data_X, i))
		else
			reconstruct(trained_vae, getobs(mnist_data_X, i))[end] .> 0.5
		end
		p̂_cvae = if trained_cvae isa Nothing
			zero(getobs(mnist_data_X, i))
		else
			reconstruct(trained_cvae, getobs(mnist_data_X, i), getobs(mnist_data_Y, i))[end] .> 0.5
		end
		p_rs = permutedims(reshape(getobs(mnist_data_X, i), 28, :), [2, 1])
		
		plt_true = plot(
			Gray.(p_rs), 
			title=i==indicies[1] ? "Ground Truth" : "", 
			axis=nothing)

		p̂_vae_rs = permutedims(reshape(p̂_vae, 28, :), [2, 1])
		plt_vae = plot(
			Gray.(p̂_vae_rs), 
			title=i==indicies[1] ? "VAE" : "", 
			axis=nothing)

		p̂_cvae_rs = permutedims(reshape(p̂_cvae, 28, :), [2, 1])
		plt_cvae = plot(
			Gray.(p̂_cvae_rs), 
			title=i==indicies[1] ? "CVAE" : "", 
			axis=nothing)
		
		push!(plts, plot(plt_true, plt_vae, plt_cvae, layout=(1,3)))
	end
	plot(plts..., layout = (length(plts), 1), size=(800, 1200))
end

# ╔═╡ 19852a30-3a3d-4be9-aa52-e30e303059ee
md"""
The CVAE and VAE can also be used to create new data. To do this we sample from our prior (i.e $\mathcal{N}(0, I)$). For the CVAE we append the appropriate class label (one hot encoded). Notice how the CVAE is able to consistently reproduce a single digit, while the VAE will produce a random set of digits!

You will need the CVAE visualization to answer some questions in the written portion of the assignment.
"""

# ╔═╡ 8a4b1faa-e8e6-49fe-9a33-8bb78fb2126e
md"
**CVAE:**
"

# ╔═╡ 39b53c97-608e-41d6-bfea-641b8247dbe9
let
	class_label = 7
	nbs = 16
	rng = Random.Xoshiro(10)
	p̂_vae = if trained_cvae isa Nothing
		randn(rng, 28^2, 16)
	else
		h = randn(rng, Float32, size(trained_cvae.encoder.μ.weight, 1), nbs)
		h_n = append_class_to_data(h, fill(class_label, nbs)) # append label to hiddenstate.
		trained_cvae.decoder(h_n)
	end
	p̂_indv = [Gray.(permutedims(reshape(p̂_vae[:, i], 28, :), [2, 1])) for i in 1:nbs]
	
	plt_vae = [plot(
		p̂_indv[i], 
		axis=nothing) for i in 1:nbs]
	plot(plt_vae...)
end

# ╔═╡ 35a70493-00ec-429d-a66a-bdf4a4abd35c
md"
**VAE:**
"

# ╔═╡ 146886ac-5a0f-4775-8984-65a1b043f0b6
let
	nbs = 16
	rng = Random.Xoshiro(10)
	p̂_vae = if trained_vae isa Nothing
		randn(rng, 28^2, 16)
	else
		h = randn(rng, Float32, size(trained_vae.encoder.μ.weight, 1), nbs)
		h_n = h
		trained_vae.decoder(h_n)
	end
	p̂_indv = [Gray.(permutedims(reshape(p̂_vae[:, i], 28, :), [2, 1])) for i in 1:nbs]
	
	plt_vae = [plot(
		p̂_indv[i], 
		axis=nothing) for i in 1:nbs]
	plot(plt_vae...)
end

# ╔═╡ 00000000-0000-0000-0000-000000000001
PLUTO_PROJECT_TOML_CONTENTS = """
[deps]
Flux = "587475ba-b771-5e3f-ad9e-33799f191a9c"
MLDatasets = "eb30cadb-4394-5ae3-aed4-317e484a6458"
MLUtils = "f1d291b0-491e-4a28-83b9-f70985020b54"
PlutoUI = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
ProgressLogging = "33c8b6b6-d38a-422a-b730-caa89a2f386c"
Random = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"
Statistics = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"
StatsPlots = "f3b207a7-027a-5e70-b257-86293d7955fd"

[compat]
Flux = "~0.13.6"
MLDatasets = "~0.7.5"
MLUtils = "~0.2.11"
PlutoUI = "~0.7.48"
ProgressLogging = "~0.1.4"
StatsPlots = "~0.15.4"
"""

# ╔═╡ 00000000-0000-0000-0000-000000000002
PLUTO_MANIFEST_TOML_CONTENTS = """
# This file is machine-generated - editing it directly is not advised

julia_version = "1.8.0"
manifest_format = "2.0"
project_hash = "9ba1df8eae08d2d0c0c42f912022973ad7f5b247"

[[deps.AbstractFFTs]]
deps = ["ChainRulesCore", "LinearAlgebra"]
git-tree-sha1 = "69f7020bd72f069c219b5e8c236c1fa90d2cb409"
uuid = "621f4979-c628-5d54-868e-fcf4e3e8185c"
version = "1.2.1"

[[deps.AbstractPlutoDingetjes]]
deps = ["Pkg"]
git-tree-sha1 = "8eaf9f1b4921132a4cff3f36a1d9ba923b14a481"
uuid = "6e696c72-6542-2067-7265-42206c756150"
version = "1.1.4"

[[deps.Accessors]]
deps = ["Compat", "CompositionsBase", "ConstructionBase", "Dates", "InverseFunctions", "LinearAlgebra", "MacroTools", "Requires", "Test"]
git-tree-sha1 = "ce67f55da3a937bb001a8d00559bdfa4dba6e4f5"
uuid = "7d9f7c33-5ae7-4f3b-8dc6-eff91059b697"
version = "0.1.20"

[[deps.Adapt]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "195c5505521008abea5aee4f96930717958eac6f"
uuid = "79e6a3ab-5dfb-504d-930d-738a2a938a0e"
version = "3.4.0"

[[deps.ArgCheck]]
git-tree-sha1 = "a3a402a35a2f7e0b87828ccabbd5ebfbebe356b4"
uuid = "dce04be8-c92d-5529-be00-80e4d2c0e197"
version = "2.3.0"

[[deps.ArgTools]]
uuid = "0dad84c5-d112-42e6-8d28-ef12dabb789f"
version = "1.1.1"

[[deps.Arpack]]
deps = ["Arpack_jll", "Libdl", "LinearAlgebra", "Logging"]
git-tree-sha1 = "91ca22c4b8437da89b030f08d71db55a379ce958"
uuid = "7d9fca2a-8960-54d3-9f78-7d1dccf2cb97"
version = "0.5.3"

[[deps.Arpack_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "JLLWrappers", "Libdl", "OpenBLAS_jll", "Pkg"]
git-tree-sha1 = "5ba6c757e8feccf03a1554dfaf3e26b3cfc7fd5e"
uuid = "68821587-b530-5797-8361-c406ea357684"
version = "3.5.1+1"

[[deps.ArrayInterface]]
deps = ["ArrayInterfaceCore", "Compat", "IfElse", "LinearAlgebra", "Static"]
git-tree-sha1 = "d6173480145eb632d6571c148d94b9d3d773820e"
uuid = "4fba245c-0d91-5ea0-9b3e-6abc04ee57a9"
version = "6.0.23"

[[deps.ArrayInterfaceCore]]
deps = ["LinearAlgebra", "SparseArrays", "SuiteSparse"]
git-tree-sha1 = "e6cba4aadba7e8a7574ab2ba2fcfb307b4c4b02a"
uuid = "30b0a656-2188-435a-8636-2ec0e6a096e2"
version = "0.1.23"

[[deps.Artifacts]]
uuid = "56f22d72-fd6d-98f1-02f0-08ddc0907c33"

[[deps.AxisAlgorithms]]
deps = ["LinearAlgebra", "Random", "SparseArrays", "WoodburyMatrices"]
git-tree-sha1 = "66771c8d21c8ff5e3a93379480a2307ac36863f7"
uuid = "13072b0f-2c55-5437-9ae7-d433b7a33950"
version = "1.0.1"

[[deps.BFloat16s]]
deps = ["LinearAlgebra", "Printf", "Random", "Test"]
git-tree-sha1 = "a598ecb0d717092b5539dbbe890c98bac842b072"
uuid = "ab4f0b2a-ad5b-11e8-123f-65d77653426b"
version = "0.2.0"

[[deps.BangBang]]
deps = ["Compat", "ConstructionBase", "Future", "InitialValues", "LinearAlgebra", "Requires", "Setfield", "Tables", "ZygoteRules"]
git-tree-sha1 = "7fe6d92c4f281cf4ca6f2fba0ce7b299742da7ca"
uuid = "198e06fe-97b7-11e9-32a5-e1d131e6ad66"
version = "0.3.37"

[[deps.Base64]]
uuid = "2a0f44e3-6c83-55bd-87e4-b1978d98bd5f"

[[deps.Baselet]]
git-tree-sha1 = "aebf55e6d7795e02ca500a689d326ac979aaf89e"
uuid = "9718e550-a3fa-408a-8086-8db961cd8217"
version = "0.1.1"

[[deps.BitFlags]]
git-tree-sha1 = "84259bb6172806304b9101094a7cc4bc6f56dbc6"
uuid = "d1d4a3ce-64b1-5f1a-9ba4-7e7e69966f35"
version = "0.1.5"

[[deps.BufferedStreams]]
git-tree-sha1 = "bb065b14d7f941b8617bc323063dbe79f55d16ea"
uuid = "e1450e63-4bb3-523b-b2a4-4ffa8c0fd77d"
version = "1.1.0"

[[deps.Bzip2_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "19a35467a82e236ff51bc17a3a44b69ef35185a2"
uuid = "6e34b625-4abd-537c-b88f-471c36dfa7a0"
version = "1.0.8+0"

[[deps.CEnum]]
git-tree-sha1 = "eb4cb44a499229b3b8426dcfb5dd85333951ff90"
uuid = "fa961155-64e5-5f13-b03f-caf6b980ea82"
version = "0.4.2"

[[deps.CSV]]
deps = ["CodecZlib", "Dates", "FilePathsBase", "InlineStrings", "Mmap", "Parsers", "PooledArrays", "SentinelArrays", "Tables", "Unicode", "WeakRefStrings"]
git-tree-sha1 = "c5fd7cd27ac4aed0acf4b73948f0110ff2a854b2"
uuid = "336ed68f-0bac-5ca0-87d4-7b16caf5d00b"
version = "0.10.7"

[[deps.CUDA]]
deps = ["AbstractFFTs", "Adapt", "BFloat16s", "CEnum", "CompilerSupportLibraries_jll", "ExprTools", "GPUArrays", "GPUCompiler", "LLVM", "LazyArtifacts", "Libdl", "LinearAlgebra", "Logging", "Printf", "Random", "Random123", "RandomNumbers", "Reexport", "Requires", "SparseArrays", "SpecialFunctions", "TimerOutputs"]
git-tree-sha1 = "49549e2c28ffb9cc77b3689dc10e46e6271e9452"
uuid = "052768ef-5323-5732-b1bb-66c8b64840ba"
version = "3.12.0"

[[deps.Cairo_jll]]
deps = ["Artifacts", "Bzip2_jll", "Fontconfig_jll", "FreeType2_jll", "Glib_jll", "JLLWrappers", "LZO_jll", "Libdl", "Pixman_jll", "Pkg", "Xorg_libXext_jll", "Xorg_libXrender_jll", "Zlib_jll", "libpng_jll"]
git-tree-sha1 = "4b859a208b2397a7a623a03449e4636bdb17bcf2"
uuid = "83423d85-b0ee-5818-9007-b63ccbeb887a"
version = "1.16.1+1"

[[deps.Calculus]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "f641eb0a4f00c343bbc32346e1217b86f3ce9dad"
uuid = "49dc2e85-a5d0-5ad3-a950-438e2897f1b9"
version = "0.5.1"

[[deps.ChainRules]]
deps = ["Adapt", "ChainRulesCore", "Compat", "Distributed", "GPUArraysCore", "IrrationalConstants", "LinearAlgebra", "Random", "RealDot", "SparseArrays", "Statistics", "StructArrays"]
git-tree-sha1 = "d7d816527558cb8373e8f7a746d88eb8a167b023"
uuid = "082447d4-558c-5d27-93f4-14fc19e9eca2"
version = "1.44.7"

[[deps.ChainRulesCore]]
deps = ["Compat", "LinearAlgebra", "SparseArrays"]
git-tree-sha1 = "e7ff6cadf743c098e08fca25c91103ee4303c9bb"
uuid = "d360d2e6-b24c-11e9-a2a3-2a2ae2dbcce4"
version = "1.15.6"

[[deps.ChangesOfVariables]]
deps = ["ChainRulesCore", "LinearAlgebra", "Test"]
git-tree-sha1 = "38f7a08f19d8810338d4f5085211c7dfa5d5bdd8"
uuid = "9e997f8a-9a97-42d5-a9f1-ce6bfc15e2c0"
version = "0.1.4"

[[deps.Clustering]]
deps = ["Distances", "LinearAlgebra", "NearestNeighbors", "Printf", "Random", "SparseArrays", "Statistics", "StatsBase"]
git-tree-sha1 = "64df3da1d2a26f4de23871cd1b6482bb68092bd5"
uuid = "aaaa29a8-35af-508c-8bc3-b662a17a0fe5"
version = "0.14.3"

[[deps.CodecZlib]]
deps = ["TranscodingStreams", "Zlib_jll"]
git-tree-sha1 = "ded953804d019afa9a3f98981d99b33e3db7b6da"
uuid = "944b1d66-785c-5afd-91f1-9de20f533193"
version = "0.7.0"

[[deps.ColorSchemes]]
deps = ["ColorTypes", "ColorVectorSpace", "Colors", "FixedPointNumbers", "Random"]
git-tree-sha1 = "1fd869cc3875b57347f7027521f561cf46d1fcd8"
uuid = "35d6a980-a343-548e-a6ea-1d62b119f2f4"
version = "3.19.0"

[[deps.ColorTypes]]
deps = ["FixedPointNumbers", "Random"]
git-tree-sha1 = "eb7f0f8307f71fac7c606984ea5fb2817275d6e4"
uuid = "3da002f7-5984-5a60-b8a6-cbb66c0b333f"
version = "0.11.4"

[[deps.ColorVectorSpace]]
deps = ["ColorTypes", "FixedPointNumbers", "LinearAlgebra", "SpecialFunctions", "Statistics", "TensorCore"]
git-tree-sha1 = "d08c20eef1f2cbc6e60fd3612ac4340b89fea322"
uuid = "c3611d14-8923-5661-9e6a-0046d554d3a4"
version = "0.9.9"

[[deps.Colors]]
deps = ["ColorTypes", "FixedPointNumbers", "Reexport"]
git-tree-sha1 = "417b0ed7b8b838aa6ca0a87aadf1bb9eb111ce40"
uuid = "5ae59095-9a9b-59fe-a467-6f913c188581"
version = "0.12.8"

[[deps.CommonSubexpressions]]
deps = ["MacroTools", "Test"]
git-tree-sha1 = "7b8a93dba8af7e3b42fecabf646260105ac373f7"
uuid = "bbf7d656-a473-5ed7-a52c-81e309532950"
version = "0.3.0"

[[deps.Compat]]
deps = ["Base64", "Dates", "DelimitedFiles", "Distributed", "InteractiveUtils", "LibGit2", "Libdl", "LinearAlgebra", "Markdown", "Mmap", "Pkg", "Printf", "REPL", "Random", "SHA", "Serialization", "SharedArrays", "Sockets", "SparseArrays", "Statistics", "Test", "UUIDs", "Unicode"]
git-tree-sha1 = "78bee250c6826e1cf805a88b7f1e86025275d208"
uuid = "34da2185-b29b-5c13-b0c7-acf172513d20"
version = "3.46.0"

[[deps.CompilerSupportLibraries_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "e66e0078-7015-5450-92f7-15fbd957f2ae"
version = "0.5.2+0"

[[deps.CompositionsBase]]
git-tree-sha1 = "455419f7e328a1a2493cabc6428d79e951349769"
uuid = "a33af91c-f02d-484b-be07-31d278c5ca2b"
version = "0.1.1"

[[deps.ConstructionBase]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "fb21ddd70a051d882a1686a5a550990bbe371a95"
uuid = "187b0558-2788-49d3-abe0-74a17ed4e7c9"
version = "1.4.1"

[[deps.ContextVariablesX]]
deps = ["Compat", "Logging", "UUIDs"]
git-tree-sha1 = "25cc3803f1030ab855e383129dcd3dc294e322cc"
uuid = "6add18c4-b38d-439d-96f6-d6bc489c04c5"
version = "0.1.3"

[[deps.Contour]]
git-tree-sha1 = "d05d9e7b7aedff4e5b51a029dced05cfb6125781"
uuid = "d38c429a-6771-53c6-b99e-75d170b6e991"
version = "0.6.2"

[[deps.Crayons]]
git-tree-sha1 = "249fe38abf76d48563e2f4556bebd215aa317e15"
uuid = "a8cc5b0e-0ffa-5ad4-8c14-923d3ee1735f"
version = "4.1.1"

[[deps.DataAPI]]
git-tree-sha1 = "46d2680e618f8abd007bce0c3026cb0c4a8f2032"
uuid = "9a962f9c-6df0-11e9-0e5d-c546b8b5ee8a"
version = "1.12.0"

[[deps.DataDeps]]
deps = ["HTTP", "Libdl", "Reexport", "SHA", "p7zip_jll"]
git-tree-sha1 = "bc0a264d3e7b3eeb0b6fc9f6481f970697f29805"
uuid = "124859b0-ceae-595e-8997-d05f6a7a8dfe"
version = "0.7.10"

[[deps.DataFrames]]
deps = ["Compat", "DataAPI", "Future", "InvertedIndices", "IteratorInterfaceExtensions", "LinearAlgebra", "Markdown", "Missings", "PooledArrays", "PrettyTables", "Printf", "REPL", "Reexport", "SortingAlgorithms", "Statistics", "TableTraits", "Tables", "Unicode"]
git-tree-sha1 = "db2a9cb664fcea7836da4b414c3278d71dd602d2"
uuid = "a93c6f00-e57d-5684-b7b6-d8193f3e46c0"
version = "1.3.6"

[[deps.DataStructures]]
deps = ["Compat", "InteractiveUtils", "OrderedCollections"]
git-tree-sha1 = "d1fff3a548102f48987a52a2e0d114fa97d730f0"
uuid = "864edb3b-99cc-5e75-8d2d-829cb0a9cfe8"
version = "0.18.13"

[[deps.DataValueInterfaces]]
git-tree-sha1 = "bfc1187b79289637fa0ef6d4436ebdfe6905cbd6"
uuid = "e2d170a0-9d28-54be-80f0-106bbe20a464"
version = "1.0.0"

[[deps.DataValues]]
deps = ["DataValueInterfaces", "Dates"]
git-tree-sha1 = "d88a19299eba280a6d062e135a43f00323ae70bf"
uuid = "e7dc6d0d-1eca-5fa6-8ad6-5aecde8b7ea5"
version = "0.4.13"

[[deps.Dates]]
deps = ["Printf"]
uuid = "ade2ca70-3891-5945-98fb-dc099432e06a"

[[deps.DefineSingletons]]
git-tree-sha1 = "0fba8b706d0178b4dc7fd44a96a92382c9065c2c"
uuid = "244e2a9f-e319-4986-a169-4d1fe445cd52"
version = "0.1.2"

[[deps.DelimitedFiles]]
deps = ["Mmap"]
uuid = "8bb1440f-4735-579b-a4ab-409b98df4dab"

[[deps.DensityInterface]]
deps = ["InverseFunctions", "Test"]
git-tree-sha1 = "80c3e8639e3353e5d2912fb3a1916b8455e2494b"
uuid = "b429d917-457f-4dbc-8f4c-0cc954292b1d"
version = "0.4.0"

[[deps.DiffResults]]
deps = ["StaticArraysCore"]
git-tree-sha1 = "782dd5f4561f5d267313f23853baaaa4c52ea621"
uuid = "163ba53b-c6d8-5494-b064-1a9d43ac40c5"
version = "1.1.0"

[[deps.DiffRules]]
deps = ["IrrationalConstants", "LogExpFunctions", "NaNMath", "Random", "SpecialFunctions"]
git-tree-sha1 = "8b7a4d23e22f5d44883671da70865ca98f2ebf9d"
uuid = "b552c78f-8df3-52c6-915a-8e097449b14b"
version = "1.12.0"

[[deps.Distances]]
deps = ["LinearAlgebra", "SparseArrays", "Statistics", "StatsAPI"]
git-tree-sha1 = "3258d0659f812acde79e8a74b11f17ac06d0ca04"
uuid = "b4f34e82-e78d-54a5-968a-f98e89d6e8f7"
version = "0.10.7"

[[deps.Distributed]]
deps = ["Random", "Serialization", "Sockets"]
uuid = "8ba89e20-285c-5b6f-9357-94700520ee1b"

[[deps.Distributions]]
deps = ["ChainRulesCore", "DensityInterface", "FillArrays", "LinearAlgebra", "PDMats", "Printf", "QuadGK", "Random", "SparseArrays", "SpecialFunctions", "Statistics", "StatsBase", "StatsFuns", "Test"]
git-tree-sha1 = "04db820ebcfc1e053bd8cbb8d8bccf0ff3ead3f7"
uuid = "31c24e10-a181-5473-b8eb-7969acd0382f"
version = "0.25.76"

[[deps.DocStringExtensions]]
deps = ["LibGit2"]
git-tree-sha1 = "c36550cb29cbe373e95b3f40486b9a4148f89ffd"
uuid = "ffbed154-4ef7-542d-bbb7-c09d3a79fcae"
version = "0.9.2"

[[deps.Downloads]]
deps = ["ArgTools", "FileWatching", "LibCURL", "NetworkOptions"]
uuid = "f43a241f-c20a-4ad4-852c-f6b1247861c6"
version = "1.6.0"

[[deps.DualNumbers]]
deps = ["Calculus", "NaNMath", "SpecialFunctions"]
git-tree-sha1 = "5837a837389fccf076445fce071c8ddaea35a566"
uuid = "fa6b7ba4-c1ee-5f82-b5fc-ecf0adba8f74"
version = "0.6.8"

[[deps.Expat_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "bad72f730e9e91c08d9427d5e8db95478a3c323d"
uuid = "2e619515-83b5-522b-bb60-26c02a35a201"
version = "2.4.8+0"

[[deps.ExprTools]]
git-tree-sha1 = "56559bbef6ca5ea0c0818fa5c90320398a6fbf8d"
uuid = "e2ba6199-217a-4e67-a87a-7c52f15ade04"
version = "0.1.8"

[[deps.FFMPEG]]
deps = ["FFMPEG_jll"]
git-tree-sha1 = "b57e3acbe22f8484b4b5ff66a7499717fe1a9cc8"
uuid = "c87230d0-a227-11e9-1b43-d7ebe4e7570a"
version = "0.4.1"

[[deps.FFMPEG_jll]]
deps = ["Artifacts", "Bzip2_jll", "FreeType2_jll", "FriBidi_jll", "JLLWrappers", "LAME_jll", "Libdl", "Ogg_jll", "OpenSSL_jll", "Opus_jll", "PCRE2_jll", "Pkg", "Zlib_jll", "libaom_jll", "libass_jll", "libfdk_aac_jll", "libvorbis_jll", "x264_jll", "x265_jll"]
git-tree-sha1 = "74faea50c1d007c85837327f6775bea60b5492dd"
uuid = "b22a6f82-2f65-5046-a5b2-351ab43fb4e5"
version = "4.4.2+2"

[[deps.FFTW]]
deps = ["AbstractFFTs", "FFTW_jll", "LinearAlgebra", "MKL_jll", "Preferences", "Reexport"]
git-tree-sha1 = "90630efff0894f8142308e334473eba54c433549"
uuid = "7a1cc6ca-52ef-59f5-83cd-3a7055c09341"
version = "1.5.0"

[[deps.FFTW_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "c6033cc3892d0ef5bb9cd29b7f2f0331ea5184ea"
uuid = "f5851436-0d7a-5f13-b9de-f02708fd171a"
version = "3.3.10+0"

[[deps.FLoops]]
deps = ["BangBang", "Compat", "FLoopsBase", "InitialValues", "JuliaVariables", "MLStyle", "Serialization", "Setfield", "Transducers"]
git-tree-sha1 = "ffb97765602e3cbe59a0589d237bf07f245a8576"
uuid = "cc61a311-1640-44b5-9fba-1b764f453329"
version = "0.2.1"

[[deps.FLoopsBase]]
deps = ["ContextVariablesX"]
git-tree-sha1 = "656f7a6859be8673bf1f35da5670246b923964f7"
uuid = "b9860ae5-e623-471e-878b-f6a53c775ea6"
version = "0.1.1"

[[deps.FileIO]]
deps = ["Pkg", "Requires", "UUIDs"]
git-tree-sha1 = "7be5f99f7d15578798f338f5433b6c432ea8037b"
uuid = "5789e2e9-d7fb-5bc7-8068-2c6fae9b9549"
version = "1.16.0"

[[deps.FilePathsBase]]
deps = ["Compat", "Dates", "Mmap", "Printf", "Test", "UUIDs"]
git-tree-sha1 = "e27c4ebe80e8699540f2d6c805cc12203b614f12"
uuid = "48062228-2e41-5def-b9a4-89aafe57970f"
version = "0.9.20"

[[deps.FileWatching]]
uuid = "7b1f6079-737a-58dc-b8bc-7a2ca5c1b5ee"

[[deps.FillArrays]]
deps = ["LinearAlgebra", "Random", "SparseArrays", "Statistics"]
git-tree-sha1 = "802bfc139833d2ba893dd9e62ba1767c88d708ae"
uuid = "1a297f60-69ca-5386-bcde-b61e274b549b"
version = "0.13.5"

[[deps.FixedPointNumbers]]
deps = ["Statistics"]
git-tree-sha1 = "335bfdceacc84c5cdf16aadc768aa5ddfc5383cc"
uuid = "53c48c17-4a7d-5ca2-90c5-79b7896eea93"
version = "0.8.4"

[[deps.Flux]]
deps = ["Adapt", "ArrayInterface", "CUDA", "ChainRulesCore", "Functors", "LinearAlgebra", "MLUtils", "MacroTools", "NNlib", "NNlibCUDA", "OneHotArrays", "Optimisers", "ProgressLogging", "Random", "Reexport", "SparseArrays", "SpecialFunctions", "Statistics", "StatsBase", "Test", "Zygote"]
git-tree-sha1 = "76ca02c7c0cb7b8337f7d2d0eadb46ed03c1e843"
uuid = "587475ba-b771-5e3f-ad9e-33799f191a9c"
version = "0.13.6"

[[deps.FoldsThreads]]
deps = ["Accessors", "FunctionWrappers", "InitialValues", "SplittablesBase", "Transducers"]
git-tree-sha1 = "eb8e1989b9028f7e0985b4268dabe94682249025"
uuid = "9c68100b-dfe1-47cf-94c8-95104e173443"
version = "0.1.1"

[[deps.Fontconfig_jll]]
deps = ["Artifacts", "Bzip2_jll", "Expat_jll", "FreeType2_jll", "JLLWrappers", "Libdl", "Libuuid_jll", "Pkg", "Zlib_jll"]
git-tree-sha1 = "21efd19106a55620a188615da6d3d06cd7f6ee03"
uuid = "a3f928ae-7b40-5064-980b-68af3947d34b"
version = "2.13.93+0"

[[deps.Formatting]]
deps = ["Printf"]
git-tree-sha1 = "8339d61043228fdd3eb658d86c926cb282ae72a8"
uuid = "59287772-0a20-5a39-b81b-1366585eb4c0"
version = "0.4.2"

[[deps.ForwardDiff]]
deps = ["CommonSubexpressions", "DiffResults", "DiffRules", "LinearAlgebra", "LogExpFunctions", "NaNMath", "Preferences", "Printf", "Random", "SpecialFunctions", "StaticArrays"]
git-tree-sha1 = "187198a4ed8ccd7b5d99c41b69c679269ea2b2d4"
uuid = "f6369f11-7733-5829-9624-2563aa707210"
version = "0.10.32"

[[deps.FreeType2_jll]]
deps = ["Artifacts", "Bzip2_jll", "JLLWrappers", "Libdl", "Pkg", "Zlib_jll"]
git-tree-sha1 = "87eb71354d8ec1a96d4a7636bd57a7347dde3ef9"
uuid = "d7e528f0-a631-5988-bf34-fe36492bcfd7"
version = "2.10.4+0"

[[deps.FriBidi_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "aa31987c2ba8704e23c6c8ba8a4f769d5d7e4f91"
uuid = "559328eb-81f9-559d-9380-de523a88c83c"
version = "1.0.10+0"

[[deps.FunctionWrappers]]
git-tree-sha1 = "d62485945ce5ae9c0c48f124a84998d755bae00e"
uuid = "069b7b12-0de2-55c6-9aab-29f3d0a68a2e"
version = "1.1.3"

[[deps.Functors]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "a2657dd0f3e8a61dbe70fc7c122038bd33790af5"
uuid = "d9f16b24-f501-4c13-a1f2-28368ffc5196"
version = "0.3.0"

[[deps.Future]]
deps = ["Random"]
uuid = "9fa8497b-333b-5362-9e8d-4d0656e87820"

[[deps.GLFW_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Libglvnd_jll", "Pkg", "Xorg_libXcursor_jll", "Xorg_libXi_jll", "Xorg_libXinerama_jll", "Xorg_libXrandr_jll"]
git-tree-sha1 = "d972031d28c8c8d9d7b41a536ad7bb0c2579caca"
uuid = "0656b61e-2033-5cc2-a64a-77c0f6c09b89"
version = "3.3.8+0"

[[deps.GPUArrays]]
deps = ["Adapt", "GPUArraysCore", "LLVM", "LinearAlgebra", "Printf", "Random", "Reexport", "Serialization", "Statistics"]
git-tree-sha1 = "45d7deaf05cbb44116ba785d147c518ab46352d7"
uuid = "0c68f7d7-f131-5f86-a1c3-88cf8149b2d7"
version = "8.5.0"

[[deps.GPUArraysCore]]
deps = ["Adapt"]
git-tree-sha1 = "6872f5ec8fd1a38880f027a26739d42dcda6691f"
uuid = "46192b85-c4d5-4398-a991-12ede77f4527"
version = "0.1.2"

[[deps.GPUCompiler]]
deps = ["ExprTools", "InteractiveUtils", "LLVM", "Libdl", "Logging", "TimerOutputs", "UUIDs"]
git-tree-sha1 = "323949b0bbdf38c93d2ea1f7d3e68ff163c3f081"
uuid = "61eb1bfa-7361-4325-ad38-22787b887f55"
version = "0.16.5"

[[deps.GR]]
deps = ["Base64", "DelimitedFiles", "GR_jll", "HTTP", "JSON", "Libdl", "LinearAlgebra", "Pkg", "Preferences", "Printf", "Random", "Serialization", "Sockets", "Test", "UUIDs"]
git-tree-sha1 = "00a9d4abadc05b9476e937a5557fcce476b9e547"
uuid = "28b8d3ca-fb5f-59d9-8090-bfdbd6d07a71"
version = "0.69.5"

[[deps.GR_jll]]
deps = ["Artifacts", "Bzip2_jll", "Cairo_jll", "FFMPEG_jll", "Fontconfig_jll", "GLFW_jll", "JLLWrappers", "JpegTurbo_jll", "Libdl", "Libtiff_jll", "Pixman_jll", "Pkg", "Qt5Base_jll", "Zlib_jll", "libpng_jll"]
git-tree-sha1 = "bc9f7725571ddb4ab2c4bc74fa397c1c5ad08943"
uuid = "d2c73de3-f751-5644-a686-071e5b155ba9"
version = "0.69.1+0"

[[deps.GZip]]
deps = ["Libdl"]
git-tree-sha1 = "039be665faf0b8ae36e089cd694233f5dee3f7d6"
uuid = "92fee26a-97fe-5a0c-ad85-20a5f3185b63"
version = "0.5.1"

[[deps.Gettext_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "JLLWrappers", "Libdl", "Libiconv_jll", "Pkg", "XML2_jll"]
git-tree-sha1 = "9b02998aba7bf074d14de89f9d37ca24a1a0b046"
uuid = "78b55507-aeef-58d4-861c-77aaff3498b1"
version = "0.21.0+0"

[[deps.Glib_jll]]
deps = ["Artifacts", "Gettext_jll", "JLLWrappers", "Libdl", "Libffi_jll", "Libiconv_jll", "Libmount_jll", "PCRE2_jll", "Pkg", "Zlib_jll"]
git-tree-sha1 = "fb83fbe02fe57f2c068013aa94bcdf6760d3a7a7"
uuid = "7746bdde-850d-59dc-9ae8-88ece973131d"
version = "2.74.0+1"

[[deps.Glob]]
git-tree-sha1 = "4df9f7e06108728ebf00a0a11edee4b29a482bb2"
uuid = "c27321d9-0574-5035-807b-f59d2c89b15c"
version = "1.3.0"

[[deps.Graphics]]
deps = ["Colors", "LinearAlgebra", "NaNMath"]
git-tree-sha1 = "d61890399bc535850c4bf08e4e0d3a7ad0f21cbd"
uuid = "a2bd30eb-e257-5431-a919-1863eab51364"
version = "1.1.2"

[[deps.Graphite2_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "344bf40dcab1073aca04aa0df4fb092f920e4011"
uuid = "3b182d85-2403-5c21-9c21-1e1f0cc25472"
version = "1.3.14+0"

[[deps.Grisu]]
git-tree-sha1 = "53bb909d1151e57e2484c3d1b53e19552b887fb2"
uuid = "42e2da0e-8278-4e71-bc24-59509adca0fe"
version = "1.0.2"

[[deps.HDF5]]
deps = ["Compat", "HDF5_jll", "Libdl", "Mmap", "Random", "Requires"]
git-tree-sha1 = "19effd6b5af759c8aaeb9c77f89422d3f975ab65"
uuid = "f67ccb44-e63f-5c2f-98bd-6dc0ccc4ba2f"
version = "0.16.12"

[[deps.HDF5_jll]]
deps = ["Artifacts", "JLLWrappers", "LibCURL_jll", "Libdl", "OpenSSL_jll", "Pkg", "Zlib_jll"]
git-tree-sha1 = "4cc2bb72df6ff40b055295fdef6d92955f9dede8"
uuid = "0234f1f7-429e-5d53-9886-15a909be8d59"
version = "1.12.2+2"

[[deps.HTTP]]
deps = ["Base64", "CodecZlib", "Dates", "IniFile", "Logging", "LoggingExtras", "MbedTLS", "NetworkOptions", "OpenSSL", "Random", "SimpleBufferStream", "Sockets", "URIs", "UUIDs"]
git-tree-sha1 = "a97d47758e933cd5fe5ea181d178936a9fc60427"
uuid = "cd3eb016-35fb-5094-929b-558a96fad6f3"
version = "1.5.1"

[[deps.HarfBuzz_jll]]
deps = ["Artifacts", "Cairo_jll", "Fontconfig_jll", "FreeType2_jll", "Glib_jll", "Graphite2_jll", "JLLWrappers", "Libdl", "Libffi_jll", "Pkg"]
git-tree-sha1 = "129acf094d168394e80ee1dc4bc06ec835e510a3"
uuid = "2e76f6c2-a576-52d4-95c1-20adfe4de566"
version = "2.8.1+1"

[[deps.HypergeometricFunctions]]
deps = ["DualNumbers", "LinearAlgebra", "OpenLibm_jll", "SpecialFunctions", "Test"]
git-tree-sha1 = "709d864e3ed6e3545230601f94e11ebc65994641"
uuid = "34004b35-14d8-5ef3-9330-4cdb6864b03a"
version = "0.3.11"

[[deps.Hyperscript]]
deps = ["Test"]
git-tree-sha1 = "8d511d5b81240fc8e6802386302675bdf47737b9"
uuid = "47d2ed2b-36de-50cf-bf87-49c2cf4b8b91"
version = "0.0.4"

[[deps.HypertextLiteral]]
deps = ["Tricks"]
git-tree-sha1 = "c47c5fa4c5308f27ccaac35504858d8914e102f9"
uuid = "ac1192a8-f4b3-4bfe-ba22-af5b92cd3ab2"
version = "0.9.4"

[[deps.IOCapture]]
deps = ["Logging", "Random"]
git-tree-sha1 = "f7be53659ab06ddc986428d3a9dcc95f6fa6705a"
uuid = "b5f81e59-6552-4d32-b1f0-c071b021bf89"
version = "0.2.2"

[[deps.IRTools]]
deps = ["InteractiveUtils", "MacroTools", "Test"]
git-tree-sha1 = "2e99184fca5eb6f075944b04c22edec29beb4778"
uuid = "7869d1d1-7146-5819-86e3-90919afe41df"
version = "0.4.7"

[[deps.IfElse]]
git-tree-sha1 = "debdd00ffef04665ccbb3e150747a77560e8fad1"
uuid = "615f187c-cbe4-4ef1-ba3b-2fcf58d6d173"
version = "0.1.1"

[[deps.ImageBase]]
deps = ["ImageCore", "Reexport"]
git-tree-sha1 = "b51bb8cae22c66d0f6357e3bcb6363145ef20835"
uuid = "c817782e-172a-44cc-b673-b171935fbb9e"
version = "0.1.5"

[[deps.ImageCore]]
deps = ["AbstractFFTs", "ColorVectorSpace", "Colors", "FixedPointNumbers", "Graphics", "MappedArrays", "MosaicViews", "OffsetArrays", "PaddedViews", "Reexport"]
git-tree-sha1 = "acf614720ef026d38400b3817614c45882d75500"
uuid = "a09fc81d-aa75-5fe9-8630-4744c3626534"
version = "0.9.4"

[[deps.ImageShow]]
deps = ["Base64", "FileIO", "ImageBase", "ImageCore", "OffsetArrays", "StackViews"]
git-tree-sha1 = "b563cf9ae75a635592fc73d3eb78b86220e55bd8"
uuid = "4e3cecfd-b093-5904-9786-8bbb286a6a31"
version = "0.3.6"

[[deps.IniFile]]
git-tree-sha1 = "f550e6e32074c939295eb5ea6de31849ac2c9625"
uuid = "83e8ac13-25f8-5344-8a64-a9f2b223428f"
version = "0.5.1"

[[deps.InitialValues]]
git-tree-sha1 = "4da0f88e9a39111c2fa3add390ab15f3a44f3ca3"
uuid = "22cec73e-a1b8-11e9-2c92-598750a2cf9c"
version = "0.3.1"

[[deps.InlineStrings]]
deps = ["Parsers"]
git-tree-sha1 = "db619c421554e1e7e07491b85a8f4b96b3f04ca0"
uuid = "842dd82b-1e85-43dc-bf29-5d0ee9dffc48"
version = "1.2.2"

[[deps.IntelOpenMP_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "d979e54b71da82f3a65b62553da4fc3d18c9004c"
uuid = "1d5cc7b8-4909-519e-a0f8-d0f5ad9712d0"
version = "2018.0.3+2"

[[deps.InteractiveUtils]]
deps = ["Markdown"]
uuid = "b77e0a4c-d291-57a0-90e8-8db25a27a240"

[[deps.InternedStrings]]
deps = ["Random", "Test"]
git-tree-sha1 = "eb05b5625bc5d821b8075a77e4c421933e20c76b"
uuid = "7d512f48-7fb1-5a58-b986-67e6dc259f01"
version = "0.7.0"

[[deps.Interpolations]]
deps = ["Adapt", "AxisAlgorithms", "ChainRulesCore", "LinearAlgebra", "OffsetArrays", "Random", "Ratios", "Requires", "SharedArrays", "SparseArrays", "StaticArrays", "WoodburyMatrices"]
git-tree-sha1 = "842dd89a6cb75e02e85fdd75c760cdc43f5d6863"
uuid = "a98d9a8b-a2ab-59e6-89dd-64a1c18fca59"
version = "0.14.6"

[[deps.InverseFunctions]]
deps = ["Test"]
git-tree-sha1 = "49510dfcb407e572524ba94aeae2fced1f3feb0f"
uuid = "3587e190-3f89-42d0-90ee-14403ec27112"
version = "0.1.8"

[[deps.InvertedIndices]]
git-tree-sha1 = "bee5f1ef5bf65df56bdd2e40447590b272a5471f"
uuid = "41ab1584-1d38-5bbf-9106-f11c6c58b48f"
version = "1.1.0"

[[deps.IrrationalConstants]]
git-tree-sha1 = "7fd44fd4ff43fc60815f8e764c0f352b83c49151"
uuid = "92d709cd-6900-40b7-9082-c6be49f344b6"
version = "0.1.1"

[[deps.IteratorInterfaceExtensions]]
git-tree-sha1 = "a3f24677c21f5bbe9d2a714f95dcd58337fb2856"
uuid = "82899510-4779-5014-852e-03e436cf321d"
version = "1.0.0"

[[deps.JLD2]]
deps = ["FileIO", "MacroTools", "Mmap", "OrderedCollections", "Pkg", "Printf", "Reexport", "TranscodingStreams", "UUIDs"]
git-tree-sha1 = "1c3ff7416cb727ebf4bab0491a56a296d7b8cf1d"
uuid = "033835bb-8acc-5ee8-8aae-3f567f8a3819"
version = "0.4.25"

[[deps.JLFzf]]
deps = ["Pipe", "REPL", "Random", "fzf_jll"]
git-tree-sha1 = "f377670cda23b6b7c1c0b3893e37451c5c1a2185"
uuid = "1019f520-868f-41f5-a6de-eb00f4b6a39c"
version = "0.1.5"

[[deps.JLLWrappers]]
deps = ["Preferences"]
git-tree-sha1 = "abc9885a7ca2052a736a600f7fa66209f96506e1"
uuid = "692b3bcd-3c85-4b1f-b108-f13ce0eb3210"
version = "1.4.1"

[[deps.JSON]]
deps = ["Dates", "Mmap", "Parsers", "Unicode"]
git-tree-sha1 = "3c837543ddb02250ef42f4738347454f95079d4e"
uuid = "682c06a0-de6a-54ab-a142-c8b1cf79cde6"
version = "0.21.3"

[[deps.JSON3]]
deps = ["Dates", "Mmap", "Parsers", "SnoopPrecompile", "StructTypes", "UUIDs"]
git-tree-sha1 = "65edf3850efb9cb4ca3b0bf488e29c6c38a23d2d"
uuid = "0f8b85d8-7281-11e9-16c2-39a750bddbf1"
version = "1.11.1"

[[deps.JpegTurbo_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "b53380851c6e6664204efb2e62cd24fa5c47e4ba"
uuid = "aacddb02-875f-59d6-b918-886e6ef4fbf8"
version = "2.1.2+0"

[[deps.JuliaVariables]]
deps = ["MLStyle", "NameResolution"]
git-tree-sha1 = "49fb3cb53362ddadb4415e9b73926d6b40709e70"
uuid = "b14d175d-62b4-44ba-8fb7-3064adc8c3ec"
version = "0.2.4"

[[deps.KernelDensity]]
deps = ["Distributions", "DocStringExtensions", "FFTW", "Interpolations", "StatsBase"]
git-tree-sha1 = "9816b296736292a80b9a3200eb7fbb57aaa3917a"
uuid = "5ab0869b-81aa-558d-bb23-cbf5423bbe9b"
version = "0.6.5"

[[deps.LAME_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "f6250b16881adf048549549fba48b1161acdac8c"
uuid = "c1c5ebd0-6772-5130-a774-d5fcae4a789d"
version = "3.100.1+0"

[[deps.LERC_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "bf36f528eec6634efc60d7ec062008f171071434"
uuid = "88015f11-f218-50d7-93a8-a6af411a945d"
version = "3.0.0+1"

[[deps.LLVM]]
deps = ["CEnum", "LLVMExtra_jll", "Libdl", "Printf", "Unicode"]
git-tree-sha1 = "e7e9184b0bf0158ac4e4aa9daf00041b5909bf1a"
uuid = "929cbde3-209d-540e-8aea-75f648917ca0"
version = "4.14.0"

[[deps.LLVMExtra_jll]]
deps = ["Artifacts", "JLLWrappers", "LazyArtifacts", "Libdl", "Pkg", "TOML"]
git-tree-sha1 = "771bfe376249626d3ca12bcd58ba243d3f961576"
uuid = "dad2f222-ce93-54a1-a47d-0025e8a3acab"
version = "0.0.16+0"

[[deps.LZO_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "e5b909bcf985c5e2605737d2ce278ed791b89be6"
uuid = "dd4b983a-f0e5-5f8d-a1b7-129d4a5fb1ac"
version = "2.10.1+0"

[[deps.LaTeXStrings]]
git-tree-sha1 = "f2355693d6778a178ade15952b7ac47a4ff97996"
uuid = "b964fa9f-0449-5b57-a5c2-d3ea65f4040f"
version = "1.3.0"

[[deps.Latexify]]
deps = ["Formatting", "InteractiveUtils", "LaTeXStrings", "MacroTools", "Markdown", "OrderedCollections", "Printf", "Requires"]
git-tree-sha1 = "ab9aa169d2160129beb241cb2750ca499b4e90e9"
uuid = "23fbe1c1-3f47-55db-b15f-69d7ec21a316"
version = "0.15.17"

[[deps.LazyArtifacts]]
deps = ["Artifacts", "Pkg"]
uuid = "4af54fe1-eca0-43a8-85a7-787d91b784e3"

[[deps.LazyModules]]
git-tree-sha1 = "a560dd966b386ac9ae60bdd3a3d3a326062d3c3e"
uuid = "8cdb02fc-e678-4876-92c5-9defec4f444e"
version = "0.3.1"

[[deps.LibCURL]]
deps = ["LibCURL_jll", "MozillaCACerts_jll"]
uuid = "b27032c2-a3e7-50c8-80cd-2d36dbcbfd21"
version = "0.6.3"

[[deps.LibCURL_jll]]
deps = ["Artifacts", "LibSSH2_jll", "Libdl", "MbedTLS_jll", "Zlib_jll", "nghttp2_jll"]
uuid = "deac9b47-8bc7-5906-a0fe-35ac56dc84c0"
version = "7.84.0+0"

[[deps.LibGit2]]
deps = ["Base64", "NetworkOptions", "Printf", "SHA"]
uuid = "76f85450-5226-5b5a-8eaa-529ad045b433"

[[deps.LibSSH2_jll]]
deps = ["Artifacts", "Libdl", "MbedTLS_jll"]
uuid = "29816b5a-b9ab-546f-933c-edad1886dfa8"
version = "1.10.2+0"

[[deps.Libdl]]
uuid = "8f399da3-3557-5675-b5ff-fb832c97cbdb"

[[deps.Libffi_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "0b4a5d71f3e5200a7dff793393e09dfc2d874290"
uuid = "e9f186c6-92d2-5b65-8a66-fee21dc1b490"
version = "3.2.2+1"

[[deps.Libgcrypt_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Libgpg_error_jll", "Pkg"]
git-tree-sha1 = "64613c82a59c120435c067c2b809fc61cf5166ae"
uuid = "d4300ac3-e22c-5743-9152-c294e39db1e4"
version = "1.8.7+0"

[[deps.Libglvnd_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libX11_jll", "Xorg_libXext_jll"]
git-tree-sha1 = "7739f837d6447403596a75d19ed01fd08d6f56bf"
uuid = "7e76a0d4-f3c7-5321-8279-8d96eeed0f29"
version = "1.3.0+3"

[[deps.Libgpg_error_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "c333716e46366857753e273ce6a69ee0945a6db9"
uuid = "7add5ba3-2f88-524e-9cd5-f83b8a55f7b8"
version = "1.42.0+0"

[[deps.Libiconv_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "42b62845d70a619f063a7da093d995ec8e15e778"
uuid = "94ce4f54-9a6c-5748-9c1c-f9c7231a4531"
version = "1.16.1+1"

[[deps.Libmount_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "9c30530bf0effd46e15e0fdcf2b8636e78cbbd73"
uuid = "4b2f31a3-9ecc-558c-b454-b3730dcb73e9"
version = "2.35.0+0"

[[deps.Libtiff_jll]]
deps = ["Artifacts", "JLLWrappers", "JpegTurbo_jll", "LERC_jll", "Libdl", "Pkg", "Zlib_jll", "Zstd_jll"]
git-tree-sha1 = "3eb79b0ca5764d4799c06699573fd8f533259713"
uuid = "89763e89-9b03-5906-acba-b20f662cd828"
version = "4.4.0+0"

[[deps.Libuuid_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "7f3efec06033682db852f8b3bc3c1d2b0a0ab066"
uuid = "38a345b3-de98-5d2b-a5d3-14cd9215e700"
version = "2.36.0+0"

[[deps.LinearAlgebra]]
deps = ["Libdl", "libblastrampoline_jll"]
uuid = "37e2e46d-f89d-539d-b4ee-838fcccc9c8e"

[[deps.LogExpFunctions]]
deps = ["ChainRulesCore", "ChangesOfVariables", "DocStringExtensions", "InverseFunctions", "IrrationalConstants", "LinearAlgebra"]
git-tree-sha1 = "94d9c52ca447e23eac0c0f074effbcd38830deb5"
uuid = "2ab3a3ac-af41-5b50-aa03-7779005ae688"
version = "0.3.18"

[[deps.Logging]]
uuid = "56ddb016-857b-54e1-b83d-db4d58db5568"

[[deps.LoggingExtras]]
deps = ["Dates", "Logging"]
git-tree-sha1 = "5d4d2d9904227b8bd66386c1138cf4d5ffa826bf"
uuid = "e6f89c97-d47a-5376-807f-9c37f3926c36"
version = "0.4.9"

[[deps.MAT]]
deps = ["BufferedStreams", "CodecZlib", "HDF5", "SparseArrays"]
git-tree-sha1 = "971be550166fe3f604d28715302b58a3f7293160"
uuid = "23992714-dd62-5051-b70f-ba57cb901cac"
version = "0.10.3"

[[deps.MIMEs]]
git-tree-sha1 = "65f28ad4b594aebe22157d6fac869786a255b7eb"
uuid = "6c6e2e6c-3030-632d-7369-2d6c69616d65"
version = "0.1.4"

[[deps.MKL_jll]]
deps = ["Artifacts", "IntelOpenMP_jll", "JLLWrappers", "LazyArtifacts", "Libdl", "Pkg"]
git-tree-sha1 = "2ce8695e1e699b68702c03402672a69f54b8aca9"
uuid = "856f044c-d86e-5d09-b602-aeab76dc8ba7"
version = "2022.2.0+0"

[[deps.MLDatasets]]
deps = ["CSV", "DataDeps", "DataFrames", "DelimitedFiles", "FileIO", "FixedPointNumbers", "GZip", "Glob", "HDF5", "ImageShow", "JLD2", "JSON3", "LazyModules", "MAT", "MLUtils", "NPZ", "Pickle", "Printf", "Requires", "SparseArrays", "Tables"]
git-tree-sha1 = "0def460f055098092043c9533e245ef77b6a8cc8"
uuid = "eb30cadb-4394-5ae3-aed4-317e484a6458"
version = "0.7.5"

[[deps.MLStyle]]
git-tree-sha1 = "0638598b2ea9c60303e036be920df8df60fe2812"
uuid = "d8e11817-5142-5d16-987a-aa16d5891078"
version = "0.4.14"

[[deps.MLUtils]]
deps = ["ChainRulesCore", "DelimitedFiles", "FLoops", "FoldsThreads", "Random", "ShowCases", "Statistics", "StatsBase", "Transducers"]
git-tree-sha1 = "824e9dfc7509cab1ec73ba77b55a916bb2905e26"
uuid = "f1d291b0-491e-4a28-83b9-f70985020b54"
version = "0.2.11"

[[deps.MacroTools]]
deps = ["Markdown", "Random"]
git-tree-sha1 = "42324d08725e200c23d4dfb549e0d5d89dede2d2"
uuid = "1914dd2f-81c6-5fcd-8719-6d5c9610ff09"
version = "0.5.10"

[[deps.MappedArrays]]
git-tree-sha1 = "e8b359ef06ec72e8c030463fe02efe5527ee5142"
uuid = "dbb5928d-eab1-5f90-85c2-b9b0edb7c900"
version = "0.4.1"

[[deps.Markdown]]
deps = ["Base64"]
uuid = "d6f4376e-aef5-505a-96c1-9c027394607a"

[[deps.MbedTLS]]
deps = ["Dates", "MbedTLS_jll", "MozillaCACerts_jll", "Random", "Sockets"]
git-tree-sha1 = "03a9b9718f5682ecb107ac9f7308991db4ce395b"
uuid = "739be429-bea8-5141-9913-cc70e7f3736d"
version = "1.1.7"

[[deps.MbedTLS_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "c8ffd9c3-330d-5841-b78e-0817d7145fa1"
version = "2.28.0+0"

[[deps.Measures]]
git-tree-sha1 = "e498ddeee6f9fdb4551ce855a46f54dbd900245f"
uuid = "442fdcdd-2543-5da2-b0f3-8c86c306513e"
version = "0.3.1"

[[deps.MicroCollections]]
deps = ["BangBang", "InitialValues", "Setfield"]
git-tree-sha1 = "4d5917a26ca33c66c8e5ca3247bd163624d35493"
uuid = "128add7d-3638-4c79-886c-908ea0c25c34"
version = "0.1.3"

[[deps.Missings]]
deps = ["DataAPI"]
git-tree-sha1 = "bf210ce90b6c9eed32d25dbcae1ebc565df2687f"
uuid = "e1d29d7a-bbdc-5cf2-9ac0-f12de2c33e28"
version = "1.0.2"

[[deps.Mmap]]
uuid = "a63ad114-7e13-5084-954f-fe012c677804"

[[deps.MosaicViews]]
deps = ["MappedArrays", "OffsetArrays", "PaddedViews", "StackViews"]
git-tree-sha1 = "b34e3bc3ca7c94914418637cb10cc4d1d80d877d"
uuid = "e94cdb99-869f-56ef-bcf0-1ae2bcbe0389"
version = "0.3.3"

[[deps.MozillaCACerts_jll]]
uuid = "14a3606d-f60d-562e-9121-12d972cd8159"
version = "2022.2.1"

[[deps.MultivariateStats]]
deps = ["Arpack", "LinearAlgebra", "SparseArrays", "Statistics", "StatsAPI", "StatsBase"]
git-tree-sha1 = "efe9c8ecab7a6311d4b91568bd6c88897822fabe"
uuid = "6f286f6a-111f-5878-ab1e-185364afe411"
version = "0.10.0"

[[deps.NNlib]]
deps = ["Adapt", "ChainRulesCore", "LinearAlgebra", "Pkg", "Requires", "Statistics"]
git-tree-sha1 = "00bcfcea7b2063807fdcab2e0ce86ef00b8b8000"
uuid = "872c559c-99b0-510c-b3b7-b6c96a88d5cd"
version = "0.8.10"

[[deps.NNlibCUDA]]
deps = ["Adapt", "CUDA", "LinearAlgebra", "NNlib", "Random", "Statistics"]
git-tree-sha1 = "4429261364c5ea5b7308aecaa10e803ace101631"
uuid = "a00861dc-f156-4864-bf3c-e6376f28a68d"
version = "0.2.4"

[[deps.NPZ]]
deps = ["Compat", "FileIO", "ZipFile"]
git-tree-sha1 = "45f77b87cb9ed5b519f31e1590258930f3b840ee"
uuid = "15e1cf62-19b3-5cfa-8e77-841668bca605"
version = "0.4.2"

[[deps.NaNMath]]
deps = ["OpenLibm_jll"]
git-tree-sha1 = "a7c3d1da1189a1c2fe843a3bfa04d18d20eb3211"
uuid = "77ba4419-2d1f-58cd-9bb1-8ffee604a2e3"
version = "1.0.1"

[[deps.NameResolution]]
deps = ["PrettyPrint"]
git-tree-sha1 = "1a0fa0e9613f46c9b8c11eee38ebb4f590013c5e"
uuid = "71a1bf82-56d0-4bbc-8a3c-48b961074391"
version = "0.1.5"

[[deps.NearestNeighbors]]
deps = ["Distances", "StaticArrays"]
git-tree-sha1 = "440165bf08bc500b8fe4a7be2dc83271a00c0716"
uuid = "b8a86587-4115-5ab1-83bc-aa920d37bbce"
version = "0.4.12"

[[deps.NetworkOptions]]
uuid = "ca575930-c2e3-43a9-ace4-1e988b2c1908"
version = "1.2.0"

[[deps.Observables]]
git-tree-sha1 = "5a9ea4b9430d511980c01e9f7173739595bbd335"
uuid = "510215fc-4207-5dde-b226-833fc4488ee2"
version = "0.5.2"

[[deps.OffsetArrays]]
deps = ["Adapt"]
git-tree-sha1 = "f71d8950b724e9ff6110fc948dff5a329f901d64"
uuid = "6fe1bfb0-de20-5000-8ca7-80f57d26f881"
version = "1.12.8"

[[deps.Ogg_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "887579a3eb005446d514ab7aeac5d1d027658b8f"
uuid = "e7412a2a-1a6e-54c0-be00-318e2571c051"
version = "1.3.5+1"

[[deps.OneHotArrays]]
deps = ["Adapt", "ChainRulesCore", "GPUArrays", "LinearAlgebra", "MLUtils", "NNlib"]
git-tree-sha1 = "2f6efe2f76d57a0ee67cb6eff49b4d02fccbd175"
uuid = "0b1bfda6-eb8a-41d2-88d8-f5af5cad476f"
version = "0.1.0"

[[deps.OpenBLAS_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "Libdl"]
uuid = "4536629a-c528-5b80-bd46-f80d51c5b363"
version = "0.3.20+0"

[[deps.OpenLibm_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "05823500-19ac-5b8b-9628-191a04bc5112"
version = "0.8.1+0"

[[deps.OpenSSL]]
deps = ["BitFlags", "Dates", "MozillaCACerts_jll", "OpenSSL_jll", "Sockets"]
git-tree-sha1 = "3c3c4a401d267b04942545b1e964a20279587fd7"
uuid = "4d8831e6-92b7-49fb-bdf8-b643e874388c"
version = "1.3.0"

[[deps.OpenSSL_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "e60321e3f2616584ff98f0a4f18d98ae6f89bbb3"
uuid = "458c3c95-2e84-50aa-8efc-19380b2a3a95"
version = "1.1.17+0"

[[deps.OpenSpecFun_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "13652491f6856acfd2db29360e1bbcd4565d04f1"
uuid = "efe28fd5-8261-553b-a9e1-b2916fc3738e"
version = "0.5.5+0"

[[deps.Optimisers]]
deps = ["ChainRulesCore", "Functors", "LinearAlgebra", "Random", "Statistics"]
git-tree-sha1 = "8a9102cb805df46fc3d6effdc2917f09b0215c0b"
uuid = "3bd65402-5787-11e9-1adc-39752487f4e2"
version = "0.2.10"

[[deps.Opus_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "51a08fb14ec28da2ec7a927c4337e4332c2a4720"
uuid = "91d4177d-7536-5919-b921-800302f37372"
version = "1.3.2+0"

[[deps.OrderedCollections]]
git-tree-sha1 = "85f8e6578bf1f9ee0d11e7bb1b1456435479d47c"
uuid = "bac558e1-5e72-5ebc-8fee-abe8a469f55d"
version = "1.4.1"

[[deps.PCRE2_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "efcefdf7-47ab-520b-bdef-62a2eaa19f15"
version = "10.40.0+0"

[[deps.PDMats]]
deps = ["LinearAlgebra", "SparseArrays", "SuiteSparse"]
git-tree-sha1 = "cf494dca75a69712a72b80bc48f59dcf3dea63ec"
uuid = "90014a1f-27ba-587c-ab20-58faa44d9150"
version = "0.11.16"

[[deps.PaddedViews]]
deps = ["OffsetArrays"]
git-tree-sha1 = "03a7a85b76381a3d04c7a1656039197e70eda03d"
uuid = "5432bcbf-9aad-5242-b902-cca2824c8663"
version = "0.5.11"

[[deps.Parsers]]
deps = ["Dates"]
git-tree-sha1 = "6c01a9b494f6d2a9fc180a08b182fcb06f0958a0"
uuid = "69de0a69-1ddd-5017-9359-2bf0b02dc9f0"
version = "2.4.2"

[[deps.Pickle]]
deps = ["DataStructures", "InternedStrings", "Serialization", "SparseArrays", "Strided", "StringEncodings", "ZipFile"]
git-tree-sha1 = "e6a34eb1dc0c498f0774bbfbbbeff2de101f4235"
uuid = "fbb45041-c46e-462f-888f-7c521cafbc2c"
version = "0.3.2"

[[deps.Pipe]]
git-tree-sha1 = "6842804e7867b115ca9de748a0cf6b364523c16d"
uuid = "b98c9c47-44ae-5843-9183-064241ee97a0"
version = "1.3.0"

[[deps.Pixman_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "b4f5d02549a10e20780a24fce72bea96b6329e29"
uuid = "30392449-352a-5448-841d-b1acce4e97dc"
version = "0.40.1+0"

[[deps.Pkg]]
deps = ["Artifacts", "Dates", "Downloads", "LibGit2", "Libdl", "Logging", "Markdown", "Printf", "REPL", "Random", "SHA", "Serialization", "TOML", "Tar", "UUIDs", "p7zip_jll"]
uuid = "44cfe95a-1eb2-52ea-b672-e2afdf69b78f"
version = "1.8.0"

[[deps.PlotThemes]]
deps = ["PlotUtils", "Statistics"]
git-tree-sha1 = "1f03a2d339f42dca4a4da149c7e15e9b896ad899"
uuid = "ccf2f8ad-2431-5c83-bf29-c5338b663b6a"
version = "3.1.0"

[[deps.PlotUtils]]
deps = ["ColorSchemes", "Colors", "Dates", "Printf", "Random", "Reexport", "SnoopPrecompile", "Statistics"]
git-tree-sha1 = "21303256d239f6b484977314674aef4bb1fe4420"
uuid = "995b91a9-d308-5afd-9ec6-746e21dbc043"
version = "1.3.1"

[[deps.Plots]]
deps = ["Base64", "Contour", "Dates", "Downloads", "FFMPEG", "FixedPointNumbers", "GR", "JLFzf", "JSON", "LaTeXStrings", "Latexify", "LinearAlgebra", "Measures", "NaNMath", "Pkg", "PlotThemes", "PlotUtils", "Printf", "REPL", "Random", "RecipesBase", "RecipesPipeline", "Reexport", "RelocatableFolders", "Requires", "Scratch", "Showoff", "SnoopPrecompile", "SparseArrays", "Statistics", "StatsBase", "UUIDs", "UnicodeFun", "Unzip"]
git-tree-sha1 = "0a56829d264eb1bc910cf7c39ac008b5bcb5a0d9"
uuid = "91a5bcdd-55d7-5caf-9e0b-520d859cae80"
version = "1.35.5"

[[deps.PlutoUI]]
deps = ["AbstractPlutoDingetjes", "Base64", "ColorTypes", "Dates", "FixedPointNumbers", "Hyperscript", "HypertextLiteral", "IOCapture", "InteractiveUtils", "JSON", "Logging", "MIMEs", "Markdown", "Random", "Reexport", "URIs", "UUIDs"]
git-tree-sha1 = "efc140104e6d0ae3e7e30d56c98c4a927154d684"
uuid = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
version = "0.7.48"

[[deps.PooledArrays]]
deps = ["DataAPI", "Future"]
git-tree-sha1 = "a6062fe4063cdafe78f4a0a81cfffb89721b30e7"
uuid = "2dfb63ee-cc39-5dd5-95bd-886bf059d720"
version = "1.4.2"

[[deps.Preferences]]
deps = ["TOML"]
git-tree-sha1 = "47e5f437cc0e7ef2ce8406ce1e7e24d44915f88d"
uuid = "21216c6a-2e73-6563-6e65-726566657250"
version = "1.3.0"

[[deps.PrettyPrint]]
git-tree-sha1 = "632eb4abab3449ab30c5e1afaa874f0b98b586e4"
uuid = "8162dcfd-2161-5ef2-ae6c-7681170c5f98"
version = "0.2.0"

[[deps.PrettyTables]]
deps = ["Crayons", "Formatting", "Markdown", "Reexport", "Tables"]
git-tree-sha1 = "dfb54c4e414caa595a1f2ed759b160f5a3ddcba5"
uuid = "08abe8d2-0d0c-5749-adfa-8a2ac140af0d"
version = "1.3.1"

[[deps.Printf]]
deps = ["Unicode"]
uuid = "de0858da-6303-5e67-8744-51eddeeeb8d7"

[[deps.ProgressLogging]]
deps = ["Logging", "SHA", "UUIDs"]
git-tree-sha1 = "80d919dee55b9c50e8d9e2da5eeafff3fe58b539"
uuid = "33c8b6b6-d38a-422a-b730-caa89a2f386c"
version = "0.1.4"

[[deps.Qt5Base_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "Fontconfig_jll", "Glib_jll", "JLLWrappers", "Libdl", "Libglvnd_jll", "OpenSSL_jll", "Pkg", "Xorg_libXext_jll", "Xorg_libxcb_jll", "Xorg_xcb_util_image_jll", "Xorg_xcb_util_keysyms_jll", "Xorg_xcb_util_renderutil_jll", "Xorg_xcb_util_wm_jll", "Zlib_jll", "xkbcommon_jll"]
git-tree-sha1 = "0c03844e2231e12fda4d0086fd7cbe4098ee8dc5"
uuid = "ea2cea3b-5b76-57ae-a6ef-0a8af62496e1"
version = "5.15.3+2"

[[deps.QuadGK]]
deps = ["DataStructures", "LinearAlgebra"]
git-tree-sha1 = "97aa253e65b784fd13e83774cadc95b38011d734"
uuid = "1fd47b50-473d-5c70-9696-f719f8f3bcdc"
version = "2.6.0"

[[deps.REPL]]
deps = ["InteractiveUtils", "Markdown", "Sockets", "Unicode"]
uuid = "3fa0cd96-eef1-5676-8a61-b3b8758bbffb"

[[deps.Random]]
deps = ["SHA", "Serialization"]
uuid = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"

[[deps.Random123]]
deps = ["Random", "RandomNumbers"]
git-tree-sha1 = "7a1a306b72cfa60634f03a911405f4e64d1b718b"
uuid = "74087812-796a-5b5d-8853-05524746bad3"
version = "1.6.0"

[[deps.RandomNumbers]]
deps = ["Random", "Requires"]
git-tree-sha1 = "043da614cc7e95c703498a491e2c21f58a2b8111"
uuid = "e6cf234a-135c-5ec9-84dd-332b85af5143"
version = "1.5.3"

[[deps.Ratios]]
deps = ["Requires"]
git-tree-sha1 = "dc84268fe0e3335a62e315a3a7cf2afa7178a734"
uuid = "c84ed2f1-dad5-54f0-aa8e-dbefe2724439"
version = "0.4.3"

[[deps.RealDot]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "9f0a1b71baaf7650f4fa8a1d168c7fb6ee41f0c9"
uuid = "c1ae055f-0cd5-4b69-90a6-9a35b1a98df9"
version = "0.1.0"

[[deps.RecipesBase]]
deps = ["SnoopPrecompile"]
git-tree-sha1 = "d12e612bba40d189cead6ff857ddb67bd2e6a387"
uuid = "3cdcf5f2-1ef4-517c-9805-6587b60abb01"
version = "1.3.1"

[[deps.RecipesPipeline]]
deps = ["Dates", "NaNMath", "PlotUtils", "RecipesBase", "SnoopPrecompile"]
git-tree-sha1 = "9b1c0c8e9188950e66fc28f40bfe0f8aac311fe0"
uuid = "01d81517-befc-4cb6-b9ec-a95719d0359c"
version = "0.6.7"

[[deps.Reexport]]
git-tree-sha1 = "45e428421666073eab6f2da5c9d310d99bb12f9b"
uuid = "189a3867-3050-52da-a836-e630ba90ab69"
version = "1.2.2"

[[deps.RelocatableFolders]]
deps = ["SHA", "Scratch"]
git-tree-sha1 = "90bc7a7c96410424509e4263e277e43250c05691"
uuid = "05181044-ff0b-4ac5-8273-598c1e38db00"
version = "1.0.0"

[[deps.Requires]]
deps = ["UUIDs"]
git-tree-sha1 = "838a3a4188e2ded87a4f9f184b4b0d78a1e91cb7"
uuid = "ae029012-a4dd-5104-9daa-d747884805df"
version = "1.3.0"

[[deps.Rmath]]
deps = ["Random", "Rmath_jll"]
git-tree-sha1 = "bf3188feca147ce108c76ad82c2792c57abe7b1f"
uuid = "79098fc4-a85e-5d69-aa6a-4863f24498fa"
version = "0.7.0"

[[deps.Rmath_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "68db32dff12bb6127bac73c209881191bf0efbb7"
uuid = "f50d1b31-88e8-58de-be2c-1cc44531875f"
version = "0.3.0+0"

[[deps.SHA]]
uuid = "ea8e919c-243c-51af-8825-aaa63cd721ce"
version = "0.7.0"

[[deps.Scratch]]
deps = ["Dates"]
git-tree-sha1 = "f94f779c94e58bf9ea243e77a37e16d9de9126bd"
uuid = "6c6a2e73-6563-6170-7368-637461726353"
version = "1.1.1"

[[deps.SentinelArrays]]
deps = ["Dates", "Random"]
git-tree-sha1 = "efd23b378ea5f2db53a55ae53d3133de4e080aa9"
uuid = "91c51154-3ec4-41a3-a24f-3f23e20d615c"
version = "1.3.16"

[[deps.Serialization]]
uuid = "9e88b42a-f829-5b0c-bbe9-9e923198166b"

[[deps.Setfield]]
deps = ["ConstructionBase", "Future", "MacroTools", "StaticArraysCore"]
git-tree-sha1 = "e2cc6d8c88613c05e1defb55170bf5ff211fbeac"
uuid = "efcf1570-3423-57d1-acb7-fd33fddbac46"
version = "1.1.1"

[[deps.SharedArrays]]
deps = ["Distributed", "Mmap", "Random", "Serialization"]
uuid = "1a1011a3-84de-559e-8e89-a11a2f7dc383"

[[deps.ShowCases]]
git-tree-sha1 = "7f534ad62ab2bd48591bdeac81994ea8c445e4a5"
uuid = "605ecd9f-84a6-4c9e-81e2-4798472b76a3"
version = "0.1.0"

[[deps.Showoff]]
deps = ["Dates", "Grisu"]
git-tree-sha1 = "91eddf657aca81df9ae6ceb20b959ae5653ad1de"
uuid = "992d4aef-0814-514b-bc4d-f2e9a6c4116f"
version = "1.0.3"

[[deps.SimpleBufferStream]]
git-tree-sha1 = "874e8867b33a00e784c8a7e4b60afe9e037b74e1"
uuid = "777ac1f9-54b0-4bf8-805c-2214025038e7"
version = "1.1.0"

[[deps.SnoopPrecompile]]
git-tree-sha1 = "f604441450a3c0569830946e5b33b78c928e1a85"
uuid = "66db9d55-30c0-4569-8b51-7e840670fc0c"
version = "1.0.1"

[[deps.Sockets]]
uuid = "6462fe0b-24de-5631-8697-dd941f90decc"

[[deps.SortingAlgorithms]]
deps = ["DataStructures"]
git-tree-sha1 = "b3363d7460f7d098ca0912c69b082f75625d7508"
uuid = "a2af1166-a08f-5f64-846c-94a0d3cef48c"
version = "1.0.1"

[[deps.SparseArrays]]
deps = ["LinearAlgebra", "Random"]
uuid = "2f01184e-e22b-5df5-ae63-d93ebab69eaf"

[[deps.SpecialFunctions]]
deps = ["ChainRulesCore", "IrrationalConstants", "LogExpFunctions", "OpenLibm_jll", "OpenSpecFun_jll"]
git-tree-sha1 = "d75bda01f8c31ebb72df80a46c88b25d1c79c56d"
uuid = "276daf66-3868-5448-9aa4-cd146d93841b"
version = "2.1.7"

[[deps.SplittablesBase]]
deps = ["Setfield", "Test"]
git-tree-sha1 = "e08a62abc517eb79667d0a29dc08a3b589516bb5"
uuid = "171d559e-b47b-412a-8079-5efa626c420e"
version = "0.1.15"

[[deps.StackViews]]
deps = ["OffsetArrays"]
git-tree-sha1 = "46e589465204cd0c08b4bd97385e4fa79a0c770c"
uuid = "cae243ae-269e-4f55-b966-ac2d0dc13c15"
version = "0.1.1"

[[deps.Static]]
deps = ["IfElse"]
git-tree-sha1 = "de4f0a4f049a4c87e4948c04acff37baf1be01a6"
uuid = "aedffcd0-7271-4cad-89d0-dc628f76c6d3"
version = "0.7.7"

[[deps.StaticArrays]]
deps = ["LinearAlgebra", "Random", "StaticArraysCore", "Statistics"]
git-tree-sha1 = "f86b3a049e5d05227b10e15dbb315c5b90f14988"
uuid = "90137ffa-7385-5640-81b9-e52037218182"
version = "1.5.9"

[[deps.StaticArraysCore]]
git-tree-sha1 = "6b7ba252635a5eff6a0b0664a41ee140a1c9e72a"
uuid = "1e83bf80-4336-4d27-bf5d-d5a4f845583c"
version = "1.4.0"

[[deps.Statistics]]
deps = ["LinearAlgebra", "SparseArrays"]
uuid = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"

[[deps.StatsAPI]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "f9af7f195fb13589dd2e2d57fdb401717d2eb1f6"
uuid = "82ae8749-77ed-4fe6-ae5f-f523153014b0"
version = "1.5.0"

[[deps.StatsBase]]
deps = ["DataAPI", "DataStructures", "LinearAlgebra", "LogExpFunctions", "Missings", "Printf", "Random", "SortingAlgorithms", "SparseArrays", "Statistics", "StatsAPI"]
git-tree-sha1 = "d1bf48bfcc554a3761a133fe3a9bb01488e06916"
uuid = "2913bbd2-ae8a-5f71-8c99-4fb6c76f3a91"
version = "0.33.21"

[[deps.StatsFuns]]
deps = ["ChainRulesCore", "HypergeometricFunctions", "InverseFunctions", "IrrationalConstants", "LogExpFunctions", "Reexport", "Rmath", "SpecialFunctions"]
git-tree-sha1 = "5783b877201a82fc0014cbf381e7e6eb130473a4"
uuid = "4c63d2b9-4356-54db-8cca-17b64c39e42c"
version = "1.0.1"

[[deps.StatsPlots]]
deps = ["AbstractFFTs", "Clustering", "DataStructures", "DataValues", "Distributions", "Interpolations", "KernelDensity", "LinearAlgebra", "MultivariateStats", "NaNMath", "Observables", "Plots", "RecipesBase", "RecipesPipeline", "Reexport", "StatsBase", "TableOperations", "Tables", "Widgets"]
git-tree-sha1 = "e0d5bc26226ab1b7648278169858adcfbd861780"
uuid = "f3b207a7-027a-5e70-b257-86293d7955fd"
version = "0.15.4"

[[deps.Strided]]
deps = ["LinearAlgebra", "TupleTools"]
git-tree-sha1 = "a7a664c91104329c88222aa20264e1a05b6ad138"
uuid = "5e0ebb24-38b0-5f93-81fe-25c709ecae67"
version = "1.2.3"

[[deps.StringEncodings]]
deps = ["Libiconv_jll"]
git-tree-sha1 = "50ccd5ddb00d19392577902f0079267a72c5ab04"
uuid = "69024149-9ee7-55f6-a4c4-859efe599b68"
version = "0.3.5"

[[deps.StructArrays]]
deps = ["Adapt", "DataAPI", "StaticArraysCore", "Tables"]
git-tree-sha1 = "13237798b407150a6d2e2bce5d793d7d9576e99e"
uuid = "09ab397b-f2b6-538f-b94a-2f83cf4a842a"
version = "0.6.13"

[[deps.StructTypes]]
deps = ["Dates", "UUIDs"]
git-tree-sha1 = "ca4bccb03acf9faaf4137a9abc1881ed1841aa70"
uuid = "856f2bd8-1eba-4b0a-8007-ebc267875bd4"
version = "1.10.0"

[[deps.SuiteSparse]]
deps = ["Libdl", "LinearAlgebra", "Serialization", "SparseArrays"]
uuid = "4607b0f0-06f3-5cda-b6b1-a6196a1729e9"

[[deps.TOML]]
deps = ["Dates"]
uuid = "fa267f1f-6049-4f14-aa54-33bafae1ed76"
version = "1.0.0"

[[deps.TableOperations]]
deps = ["SentinelArrays", "Tables", "Test"]
git-tree-sha1 = "e383c87cf2a1dc41fa30c093b2a19877c83e1bc1"
uuid = "ab02a1b2-a7df-11e8-156e-fb1833f50b87"
version = "1.2.0"

[[deps.TableTraits]]
deps = ["IteratorInterfaceExtensions"]
git-tree-sha1 = "c06b2f539df1c6efa794486abfb6ed2022561a39"
uuid = "3783bdb8-4a98-5b6b-af9a-565f29a5fe9c"
version = "1.0.1"

[[deps.Tables]]
deps = ["DataAPI", "DataValueInterfaces", "IteratorInterfaceExtensions", "LinearAlgebra", "OrderedCollections", "TableTraits", "Test"]
git-tree-sha1 = "c79322d36826aa2f4fd8ecfa96ddb47b174ac78d"
uuid = "bd369af6-aec1-5ad0-b16a-f7cc5008161c"
version = "1.10.0"

[[deps.Tar]]
deps = ["ArgTools", "SHA"]
uuid = "a4e569a6-e804-4fa4-b0f3-eef7a1d5b13e"
version = "1.10.0"

[[deps.TensorCore]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "1feb45f88d133a655e001435632f019a9a1bcdb6"
uuid = "62fd8b95-f654-4bbd-a8a5-9c27f68ccd50"
version = "0.1.1"

[[deps.Test]]
deps = ["InteractiveUtils", "Logging", "Random", "Serialization"]
uuid = "8dfed614-e22c-5e08-85e1-65c5234f0b40"

[[deps.TimerOutputs]]
deps = ["ExprTools", "Printf"]
git-tree-sha1 = "9dfcb767e17b0849d6aaf85997c98a5aea292513"
uuid = "a759f4b9-e2f1-59dc-863e-4aeb61b1ea8f"
version = "0.5.21"

[[deps.TranscodingStreams]]
deps = ["Random", "Test"]
git-tree-sha1 = "8a75929dcd3c38611db2f8d08546decb514fcadf"
uuid = "3bb67fe8-82b1-5028-8e26-92a6c54297fa"
version = "0.9.9"

[[deps.Transducers]]
deps = ["Adapt", "ArgCheck", "BangBang", "Baselet", "CompositionsBase", "DefineSingletons", "Distributed", "InitialValues", "Logging", "Markdown", "MicroCollections", "Requires", "Setfield", "SplittablesBase", "Tables"]
git-tree-sha1 = "77fea79baa5b22aeda896a8d9c6445a74500a2c2"
uuid = "28d57a85-8fef-5791-bfe6-a80928e7c999"
version = "0.4.74"

[[deps.Tricks]]
git-tree-sha1 = "6bac775f2d42a611cdfcd1fb217ee719630c4175"
uuid = "410a4b4d-49e4-4fbc-ab6d-cb71b17b3775"
version = "0.1.6"

[[deps.TupleTools]]
git-tree-sha1 = "3c712976c47707ff893cf6ba4354aa14db1d8938"
uuid = "9d95972d-f1c8-5527-a6e0-b4b365fa01f6"
version = "1.3.0"

[[deps.URIs]]
git-tree-sha1 = "e59ecc5a41b000fa94423a578d29290c7266fc10"
uuid = "5c2747f8-b7ea-4ff2-ba2e-563bfd36b1d4"
version = "1.4.0"

[[deps.UUIDs]]
deps = ["Random", "SHA"]
uuid = "cf7118a7-6976-5b1a-9a39-7adc72f591a4"

[[deps.Unicode]]
uuid = "4ec0a83e-493e-50e2-b9ac-8f72acf5a8f5"

[[deps.UnicodeFun]]
deps = ["REPL"]
git-tree-sha1 = "53915e50200959667e78a92a418594b428dffddf"
uuid = "1cfade01-22cf-5700-b092-accc4b62d6e1"
version = "0.4.1"

[[deps.Unzip]]
git-tree-sha1 = "ca0969166a028236229f63514992fc073799bb78"
uuid = "41fe7b60-77ed-43a1-b4f0-825fd5a5650d"
version = "0.2.0"

[[deps.Wayland_jll]]
deps = ["Artifacts", "Expat_jll", "JLLWrappers", "Libdl", "Libffi_jll", "Pkg", "XML2_jll"]
git-tree-sha1 = "3e61f0b86f90dacb0bc0e73a0c5a83f6a8636e23"
uuid = "a2964d1f-97da-50d4-b82a-358c7fce9d89"
version = "1.19.0+0"

[[deps.Wayland_protocols_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "4528479aa01ee1b3b4cd0e6faef0e04cf16466da"
uuid = "2381bf8a-dfd0-557d-9999-79630e7b1b91"
version = "1.25.0+0"

[[deps.WeakRefStrings]]
deps = ["DataAPI", "InlineStrings", "Parsers"]
git-tree-sha1 = "b1be2855ed9ed8eac54e5caff2afcdb442d52c23"
uuid = "ea10d353-3f73-51f8-a26c-33c1cb351aa5"
version = "1.4.2"

[[deps.Widgets]]
deps = ["Colors", "Dates", "Observables", "OrderedCollections"]
git-tree-sha1 = "fcdae142c1cfc7d89de2d11e08721d0f2f86c98a"
uuid = "cc8bc4a8-27d6-5769-a93b-9d913e69aa62"
version = "0.6.6"

[[deps.WoodburyMatrices]]
deps = ["LinearAlgebra", "SparseArrays"]
git-tree-sha1 = "de67fa59e33ad156a590055375a30b23c40299d3"
uuid = "efce3f68-66dc-5838-9240-27a6d6f5f9b6"
version = "0.5.5"

[[deps.XML2_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Libiconv_jll", "Pkg", "Zlib_jll"]
git-tree-sha1 = "58443b63fb7e465a8a7210828c91c08b92132dff"
uuid = "02c8fc9c-b97f-50b9-bbe4-9be30ff0a78a"
version = "2.9.14+0"

[[deps.XSLT_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Libgcrypt_jll", "Libgpg_error_jll", "Libiconv_jll", "Pkg", "XML2_jll", "Zlib_jll"]
git-tree-sha1 = "91844873c4085240b95e795f692c4cec4d805f8a"
uuid = "aed1982a-8fda-507f-9586-7b0439959a61"
version = "1.1.34+0"

[[deps.Xorg_libX11_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libxcb_jll", "Xorg_xtrans_jll"]
git-tree-sha1 = "5be649d550f3f4b95308bf0183b82e2582876527"
uuid = "4f6342f7-b3d2-589e-9d20-edeb45f2b2bc"
version = "1.6.9+4"

[[deps.Xorg_libXau_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "4e490d5c960c314f33885790ed410ff3a94ce67e"
uuid = "0c0b7dd1-d40b-584c-a123-a41640f87eec"
version = "1.0.9+4"

[[deps.Xorg_libXcursor_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libXfixes_jll", "Xorg_libXrender_jll"]
git-tree-sha1 = "12e0eb3bc634fa2080c1c37fccf56f7c22989afd"
uuid = "935fb764-8cf2-53bf-bb30-45bb1f8bf724"
version = "1.2.0+4"

[[deps.Xorg_libXdmcp_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "4fe47bd2247248125c428978740e18a681372dd4"
uuid = "a3789734-cfe1-5b06-b2d0-1dd0d9d62d05"
version = "1.1.3+4"

[[deps.Xorg_libXext_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libX11_jll"]
git-tree-sha1 = "b7c0aa8c376b31e4852b360222848637f481f8c3"
uuid = "1082639a-0dae-5f34-9b06-72781eeb8cb3"
version = "1.3.4+4"

[[deps.Xorg_libXfixes_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libX11_jll"]
git-tree-sha1 = "0e0dc7431e7a0587559f9294aeec269471c991a4"
uuid = "d091e8ba-531a-589c-9de9-94069b037ed8"
version = "5.0.3+4"

[[deps.Xorg_libXi_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libXext_jll", "Xorg_libXfixes_jll"]
git-tree-sha1 = "89b52bc2160aadc84d707093930ef0bffa641246"
uuid = "a51aa0fd-4e3c-5386-b890-e753decda492"
version = "1.7.10+4"

[[deps.Xorg_libXinerama_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libXext_jll"]
git-tree-sha1 = "26be8b1c342929259317d8b9f7b53bf2bb73b123"
uuid = "d1454406-59df-5ea1-beac-c340f2130bc3"
version = "1.1.4+4"

[[deps.Xorg_libXrandr_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libXext_jll", "Xorg_libXrender_jll"]
git-tree-sha1 = "34cea83cb726fb58f325887bf0612c6b3fb17631"
uuid = "ec84b674-ba8e-5d96-8ba1-2a689ba10484"
version = "1.5.2+4"

[[deps.Xorg_libXrender_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libX11_jll"]
git-tree-sha1 = "19560f30fd49f4d4efbe7002a1037f8c43d43b96"
uuid = "ea2f1a96-1ddc-540d-b46f-429655e07cfa"
version = "0.9.10+4"

[[deps.Xorg_libpthread_stubs_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "6783737e45d3c59a4a4c4091f5f88cdcf0908cbb"
uuid = "14d82f49-176c-5ed1-bb49-ad3f5cbd8c74"
version = "0.1.0+3"

[[deps.Xorg_libxcb_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "XSLT_jll", "Xorg_libXau_jll", "Xorg_libXdmcp_jll", "Xorg_libpthread_stubs_jll"]
git-tree-sha1 = "daf17f441228e7a3833846cd048892861cff16d6"
uuid = "c7cfdc94-dc32-55de-ac96-5a1b8d977c5b"
version = "1.13.0+3"

[[deps.Xorg_libxkbfile_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libX11_jll"]
git-tree-sha1 = "926af861744212db0eb001d9e40b5d16292080b2"
uuid = "cc61e674-0454-545c-8b26-ed2c68acab7a"
version = "1.1.0+4"

[[deps.Xorg_xcb_util_image_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_xcb_util_jll"]
git-tree-sha1 = "0fab0a40349ba1cba2c1da699243396ff8e94b97"
uuid = "12413925-8142-5f55-bb0e-6d7ca50bb09b"
version = "0.4.0+1"

[[deps.Xorg_xcb_util_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libxcb_jll"]
git-tree-sha1 = "e7fd7b2881fa2eaa72717420894d3938177862d1"
uuid = "2def613f-5ad1-5310-b15b-b15d46f528f5"
version = "0.4.0+1"

[[deps.Xorg_xcb_util_keysyms_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_xcb_util_jll"]
git-tree-sha1 = "d1151e2c45a544f32441a567d1690e701ec89b00"
uuid = "975044d2-76e6-5fbe-bf08-97ce7c6574c7"
version = "0.4.0+1"

[[deps.Xorg_xcb_util_renderutil_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_xcb_util_jll"]
git-tree-sha1 = "dfd7a8f38d4613b6a575253b3174dd991ca6183e"
uuid = "0d47668e-0667-5a69-a72c-f761630bfb7e"
version = "0.3.9+1"

[[deps.Xorg_xcb_util_wm_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_xcb_util_jll"]
git-tree-sha1 = "e78d10aab01a4a154142c5006ed44fd9e8e31b67"
uuid = "c22f9ab0-d5fe-5066-847c-f4bb1cd4e361"
version = "0.4.1+1"

[[deps.Xorg_xkbcomp_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_libxkbfile_jll"]
git-tree-sha1 = "4bcbf660f6c2e714f87e960a171b119d06ee163b"
uuid = "35661453-b289-5fab-8a00-3d9160c6a3a4"
version = "1.4.2+4"

[[deps.Xorg_xkeyboard_config_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Xorg_xkbcomp_jll"]
git-tree-sha1 = "5c8424f8a67c3f2209646d4425f3d415fee5931d"
uuid = "33bec58e-1273-512f-9401-5d533626f822"
version = "2.27.0+4"

[[deps.Xorg_xtrans_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "79c31e7844f6ecf779705fbc12146eb190b7d845"
uuid = "c5fb5394-a638-5e4d-96e5-b29de1b5cf10"
version = "1.4.0+3"

[[deps.ZipFile]]
deps = ["Libdl", "Printf", "Zlib_jll"]
git-tree-sha1 = "3593e69e469d2111389a9bd06bac1f3d730ac6de"
uuid = "a5390f91-8eb1-5f08-bee0-b1d1ffed6cea"
version = "0.9.4"

[[deps.Zlib_jll]]
deps = ["Libdl"]
uuid = "83775a58-1f1d-513f-b197-d71354ab007a"
version = "1.2.12+3"

[[deps.Zstd_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "e45044cd873ded54b6a5bac0eb5c971392cf1927"
uuid = "3161d3a3-bdf6-5164-811a-617609db77b4"
version = "1.5.2+0"

[[deps.Zygote]]
deps = ["AbstractFFTs", "ChainRules", "ChainRulesCore", "DiffRules", "Distributed", "FillArrays", "ForwardDiff", "GPUArrays", "GPUArraysCore", "IRTools", "InteractiveUtils", "LinearAlgebra", "LogExpFunctions", "MacroTools", "NaNMath", "Random", "Requires", "SparseArrays", "SpecialFunctions", "Statistics", "ZygoteRules"]
git-tree-sha1 = "66cc604b9a27a660e25a54e408b4371123a186a6"
uuid = "e88e6eb3-aa80-5325-afca-941959d7151f"
version = "0.6.49"

[[deps.ZygoteRules]]
deps = ["MacroTools"]
git-tree-sha1 = "8c1a8e4dfacb1fd631745552c8db35d0deb09ea0"
uuid = "700de1a5-db45-46bc-99cf-38207098b444"
version = "0.2.2"

[[deps.fzf_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "868e669ccb12ba16eaf50cb2957ee2ff61261c56"
uuid = "214eeab7-80f7-51ab-84ad-2988db7cef09"
version = "0.29.0+0"

[[deps.libaom_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "3a2ea60308f0996d26f1e5354e10c24e9ef905d4"
uuid = "a4ae2306-e953-59d6-aa16-d00cac43593b"
version = "3.4.0+0"

[[deps.libass_jll]]
deps = ["Artifacts", "Bzip2_jll", "FreeType2_jll", "FriBidi_jll", "HarfBuzz_jll", "JLLWrappers", "Libdl", "Pkg", "Zlib_jll"]
git-tree-sha1 = "5982a94fcba20f02f42ace44b9894ee2b140fe47"
uuid = "0ac62f75-1d6f-5e53-bd7c-93b484bb37c0"
version = "0.15.1+0"

[[deps.libblastrampoline_jll]]
deps = ["Artifacts", "Libdl", "OpenBLAS_jll"]
uuid = "8e850b90-86db-534c-a0d3-1478176c7d93"
version = "5.1.1+0"

[[deps.libfdk_aac_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "daacc84a041563f965be61859a36e17c4e4fcd55"
uuid = "f638f0a6-7fb0-5443-88ba-1cc74229b280"
version = "2.0.2+0"

[[deps.libpng_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Zlib_jll"]
git-tree-sha1 = "94d180a6d2b5e55e447e2d27a29ed04fe79eb30c"
uuid = "b53b4c65-9356-5827-b1ea-8c7a1a84506f"
version = "1.6.38+0"

[[deps.libvorbis_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Ogg_jll", "Pkg"]
git-tree-sha1 = "b910cb81ef3fe6e78bf6acee440bda86fd6ae00c"
uuid = "f27f6e37-5d2b-51aa-960f-b287f2bc3b7a"
version = "1.3.7+1"

[[deps.nghttp2_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "8e850ede-7688-5339-a07c-302acd2aaf8d"
version = "1.48.0+0"

[[deps.p7zip_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "3f19e933-33d8-53b3-aaab-bd5110c3b7a0"
version = "17.4.0+0"

[[deps.x264_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "4fea590b89e6ec504593146bf8b988b2c00922b2"
uuid = "1270edf5-f2f9-52d2-97e9-ab00b5d0237a"
version = "2021.5.5+0"

[[deps.x265_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "ee567a171cce03570d77ad3a43e90218e38937a9"
uuid = "dfaa095f-4041-5dcd-9319-2fabd8486b76"
version = "3.5.0+0"

[[deps.xkbcommon_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg", "Wayland_jll", "Wayland_protocols_jll", "Xorg_libxcb_jll", "Xorg_xkeyboard_config_jll"]
git-tree-sha1 = "9ebfc140cc56e8c2156a15ceac2f0302e327ac0a"
uuid = "d8fb68d0-12a3-5cfd-a85a-d49703b185fd"
version = "1.4.1+0"
"""

# ╔═╡ Cell order:
# ╟─5b40470e-6195-4307-8c78-993da5c79699
# ╠═fec77c57-c21a-42c4-8e39-3c21c804504c
# ╠═a983ed0e-fb18-4b21-985b-5214e09ee5f8
# ╠═6d0284b7-1b4e-4241-a796-034c53b06fef
# ╟─72fef447-cdf6-443a-ae6b-53c217602d98
# ╟─5897039a-9571-4648-b1f6-a74efb6aef3b
# ╠═54ff2a52-13c5-4e55-ad57-7acf5474cae4
# ╟─c69ce5c8-cd42-4ae2-aaa9-74d0daa5875f
# ╠═dfb12ef6-2968-40bc-a7eb-5284dec0c359
# ╠═913c8492-d221-4cb1-aab9-cbb2dc2026f9
# ╟─74bd4039-7a14-4c1e-95d3-e9f336293d09
# ╟─e1c3cf03-f235-4a91-aae0-7afce73272cb
# ╟─5b6b85d6-25f5-4ac9-9fe3-5e85c432f0f0
# ╟─6b7db8c9-6a3a-49ae-926a-35a8f8bfb8e8
# ╠═b37d3b82-fcc2-4ac9-94cc-86174481ba86
# ╠═31543a74-a116-4861-b153-30e6bcaecce8
# ╠═b102d8ee-1191-4580-aa53-09740cf608a4
# ╠═f9596fd6-b1fa-43d4-b84c-9a1151db372d
# ╟─0371d19e-9684-4b81-b1cb-91409fcab2d7
# ╠═fd07264f-a9c2-4dd7-83de-6fc6051986b4
# ╠═226149d2-a4ce-4713-8ff4-420c82206b07
# ╟─49288405-b370-4cfb-a0e3-01d3b9126470
# ╠═8b85f477-674e-403e-a834-36eb53ab4c30
# ╠═df07dbe6-3983-48a8-b92a-40ccd4f08a87
# ╟─54f27f56-945a-48ae-96cf-51e15ca27cbe
# ╟─f8f90dbf-3b35-4ec3-8b65-1652aab3f954
# ╠═c4df8697-e133-4e92-b271-4ba5371566fa
# ╠═697d1728-1ec3-4a76-9b9d-de2831e39214
# ╟─3d49ff71-63cf-495e-b87c-e0f3ba7b925d
# ╠═bd1e57af-6373-4815-a305-86e03f57c64e
# ╟─7b6fc4a5-a630-42e7-9924-c2862bebfdd1
# ╠═356d6228-0b23-441f-826a-ac8b1cf495e8
# ╟─f64a54e5-930d-46fd-9741-6cd87696b2e7
# ╠═36fd8579-69bb-4e6d-98a3-5e80e2a2f3b9
# ╟─5d5019a1-42fa-4076-9de1-71c575133f50
# ╠═b5cbed07-f2f0-4360-9714-5a7218831e55
# ╟─a534dcbd-33f4-4920-950b-ae3f34142bd8
# ╟─c7f2fc33-4c0d-402a-a8bf-d686458d1b67
# ╠═0390b28b-0b50-424c-8f56-4c2dd01afdc5
# ╠═46d0b27d-4cb2-4916-a226-9ab34e561cd9
# ╟─eb82338f-bfc3-4ab8-a30a-abf2ec333593
# ╠═299c202e-b1a9-4e90-8efc-b15fb44c8e5e
# ╠═6b366eac-6b94-44a8-969a-1148577b251d
# ╟─49f77c7b-6e6d-4cf9-a0ea-b90c99891570
# ╠═09d8ef7a-b96c-4f50-a470-acbbc2d5b895
# ╠═390fc91e-0fb9-4f32-9a3a-54c4df1954bf
# ╟─5b6519a7-00dd-4e18-a8e9-a9eaed60d60e
# ╠═8c41e1cf-8139-4e61-b92b-b7d201e91d2d
# ╠═38f9f22f-fabf-4278-9a5a-2f77eac7c2db
# ╠═d5fa8fc7-ddd0-4739-8ed2-d3c45995e179
# ╟─f11e0cd6-ccd9-415e-9a20-a336227945f1
# ╠═32bfe0e2-21f0-4d5e-9c60-2a2ed5ca31c3
# ╠═f43279f9-6d3c-4d16-bb31-baddb402b54c
# ╠═131a7831-250a-47e0-90f7-9cb14db0a53b
# ╠═e6283131-2e8b-453f-867c-56cd91a48527
# ╠═4734248c-4542-4531-9b43-ee14aa10f306
# ╟─17db8cf5-9c65-4445-b629-4b96dbd3a319
# ╠═0df361d4-5a02-4445-b792-776080ec0ae6
# ╟─fc14bedb-e3df-4c2d-9902-1b9621e372be
# ╠═e417b47c-9f48-4d3c-82be-0a890c306266
# ╠═b76e5669-7ed6-4006-9007-934136bd9350
# ╟─7402bdc2-eb6f-4837-96f9-2d113d826b55
# ╠═91aad47c-b715-42b2-8f79-074b016356c3
# ╠═a1e5e164-68f7-4838-8a06-d268c263fdc7
# ╠═557c4bc1-74e7-409a-b620-2bb6592c5ee9
# ╠═dab0476f-8da9-4d6e-a418-9c4e0eda9e4a
# ╠═b1a674fa-8893-4b6a-b38e-a71474165786
# ╠═0616c043-10e3-4d1e-8328-7a694a931dce
# ╟─b3d2988a-d99b-4b16-bace-31f0134ed36b
# ╠═e32e185d-bfc9-411a-ade0-c6cb843ef0d6
# ╟─d1fd5e60-bb94-447e-9b35-39ef07d328b9
# ╟─0ac7b433-e358-4bb8-99de-bf43d6a5cf23
# ╠═de0817a7-0106-416e-a8de-dddcc99c3a5e
# ╠═60abff0f-1a7a-47a4-94b3-880a5894d139
# ╟─529abd48-aa20-445d-ba5d-44b31749de70
# ╠═25215b53-9a20-4cbc-bcee-17ed057950aa
# ╠═dec6630f-504b-44dd-9e0a-0f9e796691ff
# ╟─a70afce3-8dab-4b1b-a09b-c81cc45cd09d
# ╠═de0ba0ba-c061-4cae-8198-79524d892487
# ╠═ef533aa8-eccb-46a4-8ff9-ca49a233a46e
# ╟─213781b6-0ca3-4439-908d-b1c1895be24e
# ╟─86b19883-6574-4f10-a822-4186c4c0235c
# ╠═33eacfac-60bd-4b89-afd3-1e92b61cf55d
# ╠═4c5d66c1-aa1e-49e9-b004-ac66fbc783c4
# ╟─64fa9dbe-f9f4-407d-8e19-02831a097eb6
# ╠═e31da691-aad0-496c-addc-6730763ee7bb
# ╟─5497634b-361e-4ad1-b800-215b91426af2
# ╠═acea195c-0075-4217-8235-6e444d2024a8
# ╠═c0eff92e-b378-4da1-b44a-02fce5095612
# ╟─08cc03f4-9338-422f-9884-aa2745d67812
# ╟─3b720427-6fb8-46e9-8cbe-93d69c0b268d
# ╠═e6cdb6d4-6c4a-473b-85cc-b45f7de2409c
# ╟─19852a30-3a3d-4be9-aa52-e30e303059ee
# ╟─8a4b1faa-e8e6-49fe-9a33-8bb78fb2126e
# ╠═39b53c97-608e-41d6-bfea-641b8247dbe9
# ╟─35a70493-00ec-429d-a66a-bdf4a4abd35c
# ╠═146886ac-5a0f-4775-8984-65a1b043f0b6
# ╟─00000000-0000-0000-0000-000000000001
# ╟─00000000-0000-0000-0000-000000000002
